from enum import Enum


class Action(str, Enum):
    """ An Action is a required part of a Call message. """
    Authorize = "Authorize"
    BootNotification = "BootNotification"
    CancelReservation = "CancelReservation"
    CertificateSigned = "CertificateSigned"
    ChangeAvailability = "ChangeAvailability"
    ClearCache = "ClearCache"
    ClearChargingProfile = "ClearChargingProfile"
    ClearDisplayMessage = "ClearDisplayMessage"
    ClearedChargingLimit = "ClearedChargingLimit"
    ClearVariableMonitoring = "ClearVariableMonitoring"
    CostUpdate = "CostUpdate"
    CustomerInformation = "CustomerInformation"
    DataTransfer = "DataTransfer"
    DeleteCertificate = "DeleteCertificate"
    FirmwareStatusNotification = "FirmwareStatusNotification"
    Get15118EVCertificate = "Get15118EVCertificate"
    GetBaseReport = "GetBaseReport"
    GetCertificateStatus = "GetCertificateStatus"
    GetChargingProfiles = "GetChargingProfiles"
    GetCompositeSchedule = "GetCompositeSchedule"
    GetDisplayMessages = "GetDisplayMessages"
    GetInstalledCertificateIds = "GetInstalledCertificateIds"
    GetLocalListVersion = "GetLocalListVersion"
    GetLog = "GetLog"
    GetMonitoringReport = "GetMonitoringReport"
    GetReport = "GetReport"
    GetTransactionStatus = "GetTransactionStatus"
    GetVariables = "GetVariables"
    Heartbeat = "Heartbeat"
    InstallCertificate = "InstallCertificate"
    LogStatusNotification = "LogStatusNotification"
    MeterValues = "MeterValues"
    NotifyChargingLimit = "NotifyChargingLimit"
    NotifyCustomerInformation = "NotifyCustomerInformation"
    NotifyDisplayMessages = "NotifyDisplayMessages"
    NotifyEVChargingNeeds = "NotifyEVChargingNeeds"
    NotifyEVChargingSchedule = "NotifyEVChargingSchedule"
    NotifyEvent = "NotifyEvent"
    NotifyMonitoringReport = "NotifyMonitoringReport"
    NotifyReport = "NotifyReport"
    PublishFirmware = "PublishFirmware"
    PublishFirmwareStatusNotification = "PublishFirmwareStatusNotification"
    ReportChargingProfiles = "ReportChargingProfiles"
    RequestStartTransaction = "RequestStartTransaction"
    RequestStopTransaction = "RequestStopTransaction"
    ReservationStatusUpdate = "ReservationStatusUpdate"
    ReserveNow = "ReserveNow"
    Reset = "Reset"
    SecurityEventNotification = "SecurityEventNotification"
    SendLocalList = "SendLocalList"
    SetChargingProfile = "SetChargingProfile"
    SetDisplayMessage = "SetDisplayMessage"
    SetMonitoringBase = "SetMonitorBase"
    SetMonitoringLevel = "SetMonitoringLevel"
    SetNetworkProfile = "SetNetworkProfile"
    SetVariableMonitoring = "SetVariableMonitoring"
    SetVariables = "SetVariables"
    SignCertificate = "SignCertificate"
    StatusNotification = "StatusNotification"
    TransactionEvent = "TransactionEvent"
    TriggerMessage = "TriggerMessage"
    UnlockConnector = "UnlockConnector"
    UnpublishFirmware = "UnpublishFirmware"
    UpdateFirmware = "UpdateFirmware"

# Enums


class APNAuthenticationType(str, Enum):
    """
    APNAuthenticationEnumType is used by
    setNetworkProfileSetNetworkProfileRequest.APNType
    """
    chap = "CHAP"
    none = "NONE"
    pap = "PAP"
    auto = "AUTO"


class AttributeType(str, Enum):
    """
    AttributeEnumType is used by Common:VariableAttributeType,
    getVariables:GetVariablesRequest.GetVariableDataType ,
    getVariables:GetVariablesResponse.GetVariableResultType ,
    setVariables:SetVariablesRequest.SetVariableDataType ,
    setVariables:SetVariablesResponse.SetVariableResultType
    """
    actual = "Actual"
    target = "Target"
    minSet = "MinSet"
    maxSet = "MaxSet"


class AuthorizationStatusType(str, Enum):
    """
    Elements that constitute an entry of a Local Authorization List update.
    """

    accepted = "Accepted"
    blocked = "Blocked"
    concurrent_tx = "ConcurrentTx"
    expired = "Expired"
    invalid = "Invalid"
    # Identifier is valid, but EV Driver doesn’t have enough credit to start
    # charging. Not allowed for charging.
    no_credit = "NoCredit"
    # Identifier is valid, but not allowed to charge in this type of EVSE.
    not_allowed_type_evse = "NotAllowedTypeEVSE"
    not_at_this_location = "NotAtThisLocation"
    not_at_this_time = "NotAtThisTime"
    unknown = "Unknown"


class AuthorizeCertificateStatusType(str, Enum):
    """
    Status of the EV Contract certificate.
    """

    accepted = "Accepted"
    signature_error = "SignatureError"
    certificate_expired = "CertificateExpired"
    certificate_revoked = "CertificateRevoked"
    no_certificate_available = "NoCertificateAvailable"
    cert_chain_error = "CertChainError"
    contract_cancelled = "ContractCancelled"


class BootReasonType(str, Enum):
    """
    BootReasonEnumType is used by bootNotificationBootNotificationRequest
    """

    application_reset = "ApplicationReset"
    firmware_update = "FirmwareUpdate"
    local_reset = "LocalReset"
    power_up = "PowerUp"
    remote_reset = "RemoteReset"
    scheduled_reset = "ScheduledReset"
    triggered = "Triggered"
    unknown = "Unknown"
    watchdog = "Watchdog"


class CancelReservationStatusType(str, Enum):
    """
    Status in CancelReservationResponse.
    """

    accepted = "Accepted"
    rejected = "Rejected"


class CertificateActionType(str, Enum):
    """
    CertificateActionEnumType is used by
    get15118EVCertificateGet15118EVCertificateRequest
    """
    install = "Install"
    update = "Update"


class CertificateSignedStatusType(str, Enum):
    accepted = "Accepted"
    rejected = "Rejected"


class CertificateSigningUseType(str, Enum):
    """
    CertificateSigningUseEnumType is used by signCertificate
    SignCertificateRequest ,
    certificateSignedCertificateSignedRequest
    """
    charging_station_certificate = "ChargingStationCertificate"
    v2g_certificate = "V2GCertificate"


class ChangeAvailabilityStatusType(str, Enum):
    """
    Status returned in response to ChangeAvailability.req.
    """

    accepted = "Accepted"
    rejected = "Rejected"
    scheduled = "Scheduled"


class ChargingLimitSourceType(str, Enum):
    """
    Enumeration for indicating from which source a charging limit originates.
    """

    ems = "EMS"
    other = "Other"
    so = "SO"
    cso = "CSO"


class ChargingProfileKindType(str, Enum):
    """
    "Absolute" Schedule periods are relative to a fixed point in time defined
                in the schedule.
    "Recurring" Schedule restarts periodically at the first schedule period.
    "Relative" Schedule periods are relative to a situation- specific start
                point(such as the start of a session)
    """

    absolute = "Absolute"
    recurring = "Recurring"
    relative = "Relative"


class ChargingProfilePurposeType(str, Enum):
    """
    In load balancing scenarios, the Charge Point has one or more local
    charging profiles that limit the power or current to be shared by all
    connectors of the Charge Point. The Central System SHALL configure such
    a profile with ChargingProfilePurpose set to “ChargePointMaxProfile”.
    ChargePointMaxProfile can only be set at Charge Point ConnectorId 0.

    Default schedules for new transactions MAY be used to impose charging
    policies. An example could be a policy that prevents charging during
    the day. For schedules of this purpose, ChargingProfilePurpose SHALL
    be set to TxDefaultProfile. If TxDefaultProfile is set to ConnectorId 0,
    the TxDefaultProfile is applicable to all Connectors. If ConnectorId is
    set >0, it only applies to that specific connector. In the event a
    TxDefaultProfile for connector 0 is installed, and the Central
    System sends a new profile with ConnectorId >0, the TxDefaultProfile
    SHALL be replaced only for that specific connector.

    If a transaction-specific profile with purpose TxProfile is present,
    it SHALL overrule the default charging profile with purpose
    TxDefaultProfile for the duration of the current transaction only.
    After the transaction is stopped, the profile SHOULD be deleted.
    If there is no transaction active on the connector specified in a
    charging profile of type TxProfile, then the Charge Point SHALL
    discard it and return an error status in SetChargingProfileResponse.
    TxProfile SHALL only be set at Charge Point ConnectorId >0.

    It is not possible to set a ChargingProfile with purpose set to
    TxProfile without presence of an active transaction, or in advance of
    a transaction.

    In order to ensure that the updated charging profile applies only to the
    current transaction, the chargingProfilePurpose of the ChargingProfile
    MUST be set to TxProfile.
    """
    charging_station_external_constraints = "ChargingStationExternalConstraints"  # noqa: E501
    charging_station_max_profile = "ChargingStationMaxProfile"
    tx_default_profile = "TxDefaultProfile"
    tx_profile = "TxProfile"


class ChargingProfileStatus(str, Enum):
    """
    Status returned in response to SetChargingProfile.req.
    """

    accepted = "Accepted"
    rejected = "Rejected"


class ChargingRateUnitType(str, Enum):
    """
    Unit in which a charging schedule is defined, as used in
    GetCompositeSchedule.req and ChargingSchedule
    """

    watts = "W"
    amps = "A"


class ChargingStateType(str, Enum):
    """
    The state of the charging process.
    """

    charging = "Charging"
    ev_connected = "EVConnected"
    suspended_ev = "SuspendedEV"
    suspended_evse = "SuspendedEVSE"
    idle = "Idle"


class ClearCacheStatusType(str, Enum):
    """
    Status returned in response to ClearCache.req.
    """

    accepted = "Accepted"
    rejected = "Rejected"


class ClearChargingProfileStatusType(str, Enum):
    """
    Status returned in response to ClearChargingProfile.req.
    """

    accepted = "Accepted"
    unknown = "Unknown"


class ClearMessageStatusType(str, Enum):
    """
    Status returned in response to ClearDisplayMessageRequest.
    """

    accepted = "Accepted"
    unknown = "Unknown"


class ClearMonitoringStatusType(str, Enum):
    """
    ClearMonitoringStatusEnumType is used by CommonClearMonitoringResultType
    """

    accepted = "Accepted"
    rejected = "Rejected"
    not_found = "NotFound"


class ComponentCriterionType(str, Enum):
    """
    ComponentCriterionEnumType is used by getReportGetReportRequest
    """

    active = "Active"
    available = "Available"
    enabled = "Enabled"
    problem = "Problem"


class ConnectorStatusType(str, Enum):
    """
    Status reported in StatusNotification.req. A status can be reported for
    the Charge Point main controller (connectorId = 0) or for a specific
    connector. Status for the Charge Point main controller is a subset of the
    enumeration Available, Unavailable or Faulted.

    States considered Operative are Available, Preparing, Charging,
    SuspendedEVSE, SuspendedEV, Finishing, Reserved.
    States considered Inoperative are Unavailable, Faulted.
    """

    available = "Available"
    occupied = "Occupied"
    reserved = "Reserved"
    unavailable = "Unavailable"
    faulted = "Faulted"


class ConnectorType(str, Enum):
    """
    Allowed values of ConnectorCode.
    """
    # Combined Charging System 1 (captive cabled) a.k.a. Combo 1
    c_ccs1 = "cCCS1"
    # Combined Charging System 2 (captive cabled) a.k.a. Combo 2
    c_ccs2 = "cCCS2"
    # JARI G105-1993 (captive cabled) a.k.a. CHAdeMO
    c_g105 = "cG105"
    # Tesla Connector (captive cabled)
    c_tesla = "cTesla"
    # IEC62196-2 Type 1 connector (captive cabled) a.k.a. J1772
    c_type1 = "cType1"
    # IEC62196-2 Type 2 connector (captive cabled) a.k.a. Mennekes connector
    c_type2 = "cType2"
    # 16A 1 phase IEC60309 socket
    s309_1p_16a = "s309-1P-16A"
    s309_1p_32a = "s309-1P-32A"
    s309_3p_16a = "s309-3P-16A"
    s309_3p_32a = "s309-3P-32A"
    s_bs1361 = "sBS1361"
    s_cee_7_7 = "sCEE-7-7"
    s_type2 = "sType2"
    s_type3 = "sType3"
    other_1ph_max_16a = "Other1PhMax16A"
    other_1ph_over_16a = "Other1PhOver16A"
    other_3ph = "Other3Ph"
    pan = "Pan"
    w_inductive = "wInductive"
    w_resonant = "wResonant"
    undetermined = "Undetermined"
    unknown = "Unknown"


class CostKindType(str, Enum):
    """
    CostKindEnumType is used by CommonCostType
    """

    carbon_dioxide_emission = "CarbonDioxideEmission"
    relative_price_percentage = "RelativePricePercentage"
    renewable_generation_percentage = "RenewableGenerationPercentage"


class CustomerInformationStatusType(str, Enum):
    """
    Status in CustomerInformationResponse
    """

    accepted = "Accepted"
    rejected = "Rejected"
    invalid = "Invalid"


class DataTransferStatusType(str, Enum):
    """
    Status in DataTransferResponse.
    """
    accepted = "Accepted"
    rejected = "Rejected"
    unknown_message_id = "UnknownMessageId"
    unknown_vendor_id = "UnknownVendorId"


class DataType(str, Enum):
    """
    DataEnumType is used by CommonVariableCharacteristicsType
    """
    string = "string"
    decimal = "decimal"
    integer = "integer"
    date_time = "dateTime"
    boolean = "boolean"
    option_list = "OptionList"
    sequence_list = "SequenceList"
    member_list = "MemberList"


class DeleteCertificateStatusType(str, Enum):
    """
    DeleteCertificateStatusEnumType is used by
    deleteCertificateDeleteCertificateResponse
    """
    accepted = "Accepted"
    failed = "Failed"
    not_found = "NotFound"


class DisplayMessageStatusType(str, Enum):
    """
    Result for a SetDisplayMessageRequest as used in a
    SetDisplayMessageResponse.
    """
    accepted = "Accepted"
    not_supported_message_format = "NotSupportedMessageFormat"
    rejected = "Rejected"
    not_supported_priority = "NotSupportedPriority"
    not_supported_state = "NotSupportedState"
    unknown_transaction = "UnknownTransaction"


class EnergyTransferModeType(str, Enum):
    """
    Enumeration of energy transfer modes.
    """
    dc = "DC"
    ac_single_phase = "AC_single_phase"
    ac_two_phase = "AC_two_phase"
    ac_three_phase = "AC_three_phase"


class EventNotificationType(str, Enum):
    """
    Specifies the event notification type of the message.
    """
    hard_wired_notification = "HardWiredNotification"
    hard_wired_monitor = "HardWiredMonitor"
    preconfigured_monitor = "PreconfiguredMonitor"
    custom_monitor = "CustomMonitor"


class EventTriggerType(str, Enum):
    """
    EventTriggerEnumType is used by
    notifyEventNotifyEventRequest.EventDataType
    """
    alerting = "Alerting"
    delta = "Delta"
    periodic = "Periodic"


class FirmwareStatusType(str, Enum):
    """
    Status of a firmware download as reported in
    FirmwareStatusNotificationRequest
    """
    downloaded = "Downloaded"
    download_failed = "DownloadFailed"
    downloading = "Downloading"
    download_scheduled = "DownloadScheduled"
    download_paused = "DownloadPaused"
    idle = "Idle"
    installation_failed = "InstallationFailed"
    installing = "Installing"
    installed = "Installed"
    install_rebooting = "InstallRebooting"
    install_scheduled = "InstallScheduled"
    install_verification_failed = "InstallVerificationFailed"
    invalid_signature = "InvalidSignature"
    signature_verified = "SignatureVerified"


class GenericDeviceModelStatusType(str, Enum):
    """
    Status of a firmware download as reported in GetBaseReportResponse
    """
    accepted = "Accepted"
    rejected = "Rejected"
    not_supported = "NotSupported"
    empty_result_set = "EmptyResultSet"


class GenericStatusType(str, Enum):
    """
    Generic message response status
    """
    accepted = "Accepted"
    rejected = "Rejected"


class GetCertificateIdUseType(str, Enum):
    v2g_root_certificate = "V2GRootCertificate"
    mo_root_certificate = "MORootCertificate"
    csms_root_certificate = "CSMSRootCertificate"
    v2g_certificate_chain = "V2GCertificateChain"
    manufacturer_root_certificate = "ManufacturerRootCertificate"


class GetCertificateStatusType(str, Enum):
    """
    GetCertificateStatusEnumType is used by
     getCertificateStatusGetCertificateStatusResponse
    """
    accepted = "Accepted"
    failed = "Failed"


class GetChargingProfileStatusType(str, Enum):
    """
    GetChargingProfileStatusEnumType is used by
    getChargingProfilesGetChargingProfilesResponse
    """
    accepted = "Accepted"
    no_profiles = "NoProfiles"


class GetDisplayMessagesStatusType(str, Enum):
    """
    GetDisplayMessagesStatusEnumType is used by
    getDisplayMessagesGetDisplayMessagesResponse
    """
    accepted = "Accepted"
    unknown = "Unknown"


class GetInstalledCertificateStatusType(str, Enum):
    """
    GetInstalledCertificateStatusEnumType is used by
    getInstalledCertificateIdsGetInstalledCertificateIdsResponse
    """
    accepted = "Accepted"
    notFound = "NotFound"


class GetVariableStatusType(str, Enum):
    """
    GetVariableStatusEnumType is used by
    getVariablesGetVariablesResponse.GetVariableResultType
    """
    accepted = "Accepted"
    rejected = "Rejected"
    unknown_component = "UnknownComponent"
    unknown_variable = "UnknownVariable"
    not_supported_attribute_type = "NotSupportedAttributeType"


class HashAlgorithmType(str, Enum):
    """
    HashAlgorithmEnumType is used by
    CommonCertificateHashDataType , CommonOCSPRequestDataType
    """
    sha256 = "SHA256"
    sha384 = "SHA384"
    sha512 = "SHA512"


class IdTokenType(str, Enum):
    """
    Allowable values of the IdTokenType field.
    """
    central = "Central"
    e_maid = "eMAID"
    iso14443 = "ISO14443"
    iso15693 = "ISO15693"
    key_code = "KeyCode"
    local = "Local"
    mac_address = "MacAddress"
    no_authorization = "NoAuthorization"


class InstallCertificateStatusType(str, Enum):
    """
    InstallCertificateStatusEnumType is used by
    installCertificateInstallCertificateResponse
    """
    accepted = "Accepted"
    rejected = "Rejected"
    failed = "Failed"


class InstallCertificateUseType(str, Enum):
    """
    InstallCertificateUseEnumType is used by
    installCertificateInstallCertificateRequest
    """
    v2g_root_certificate = "V2GRootCertificate"
    mo_root_certificate = "MORootCertificate"
    csms_root_certificate = "CSMSRootCertificate"
    manufacturer_root_certificate = "ManufacturerRootCertificate"


class Iso15118EVCertificateStatusType(str, Enum):
    """
    Iso15118EVCertificateStatusEnumType is used by
    get15118EVCertificateGet15118EVCertificateResponse
    """
    accepted = "Accepted"
    failed = "Failed"


class LocationType(str, Enum):
    """
    Allowable values of the optional "location" field of a value element in
    SampledValue.
    """
    body = "Body"
    cable = "Cable"
    ev = "EV"
    inlet = "Inlet"
    outlet = "Outlet"


class LogType(str, Enum):
    """
    LogEnumType is used by getLogGetLogRequest
    """
    diagnostics_log = "DiagnosticsLog"
    security_log = "SecurityLog"


class LogStatusType(str, Enum):
    """
    Generic message response status
    """

    accepted = "Accepted"
    rejected = "Rejected"
    accepted_canceled = "AcceptedCanceled"


class MeasurandType(str, Enum):
    """
    Allowable values of the optional "measurand" field of a Value element, as
    used in MeterValues.req and StopTransaction.req messages. Default value of
    "measurand" is always "Energy.Active.Import.Register"
    """

    current_export = "Current.Export"
    current_import = "Current.Import"
    current_offered = "Current.Offered"
    energy_active_export_register = "Energy.Active.Export.Register"
    energy_active_import_register = "Energy.Active.Import.Register"
    energy_reactive_export_register = "Energy.Reactive.Export.Register"
    energy_reactive_import_register = "Energy.Reactive.Import.Register"
    energy_active_export_interval = "Energy.Active.Export.Interval"
    energy_active_import_interval = "Energy.Active.Import.Interval"
    energy_active_net = "Energy.Active.Net"
    energy_reactive_export_interval = "Energy.Reactive.Export.Interval"
    energy_reactive_import_interval = "Energy.Reactive.Import.Interval"
    energy_reactive_net = "Energy.Reactive.Net"
    energy_apparent_net = "Energy.Apparent.Net"
    energy_apparent_import = "Energy.Apparent.Import"
    energy_apparent_export = "Energy.Apparent.Export"
    frequency = "Frequency"
    power_active_export = "Power.Active.Export"
    power_active_import = "Power.Active.Import"
    power_factor = "Power.Factor"
    power_offered = "Power.Offered"
    power_reactive_export = "Power.Reactive.Export"
    power_reactive_import = "Power.Reactive.Import"
    soc = "SoC"
    voltage = "Voltage"


class MessageFormatType(str, Enum):
    """
    Format of a message to be displayed on the display of the Charging Station.
    """
    ascii = "ASCII"
    html = "HTML"
    uri = "URI"
    utf8 = "UTF8"


class MessagePriorityType(str, Enum):
    """
    Priority with which a message should be displayed on a Charging Station.
    """
    always_front = "AlwaysFront"
    in_front = "InFront"
    normal_cycle = "NormalCycle"


class MessageStateType(str, Enum):
    """
    State of the Charging Station during which a message SHALL be displayed.
    """
    charging = "Charging"
    faulted = "Faulted"
    idle = "Idle"


class MessageTriggerType(str, Enum):
    """
    Type of request to be triggered in a TriggerMessage.req
    """

    boot_notification = "BootNotification"
    log_status_notification = "LogStatusNotification"
    firmware_status_notification = "FirmwareStatusNotification"
    heartbeat = "Heartbeat"
    meter_values = "MeterValues"
    # Triggers a SignCertificate with typeOfCertificate
    # ChargingStationCertificate.
    sign_charging_station_certificate = "SignChargingStationCertificate"
    # Triggers a SignCertificate with typeOfCertificate V2GCertificate
    sign_v2_gcertificate = "SignV2GCertificate"
    status_notification = "StatusNotification"
    transaction_event = "TransactionEvent"
    sign_combined_certificate = "SignCombinedCertificate"
    publish_firmware_status_notification = "PublishFirmwareStatusNotification"


class MonitorType(str, Enum):
    """
    MonitorEnumType is used by CommonVariableMonitoringType
    """
    upper_threshold = "UpperThreshold"
    lower_threshold = "LowerThreshold"
    delta = "Delta"
    periodic = "Periodic"
    periodic_clock_aligned = "PeriodicClockAligned"


class MonitorBaseType(str, Enum):
    """
    MonitoringBaseEnumType is used by
    setMonitoringBaseSetMonitoringBaseRequest
    """
    all = "All"
    factory_default = "FactoryDefault"
    hard_wired_only = "HardWiredOnly"


class MonitoringCriterionType(str, Enum):
    """
    MonitoringCriterionEnumType is used by
    getMonitoringReportGetMonitoringReportRequest
    """
    threshold_monitoring = "ThresholdMonitoring"
    delta_monitoring = "DeltaMonitoring"
    periodic_monitoring = "PeriodicMonitoring"


class MutabilityType(str, Enum):
    """
    MutabilityEnumType is used by CommonVariableAttributeType
    """
    read_only = "ReadOnly"
    write_only = "WriteOnly"
    read_write = "ReadWrite"


class NotifyEVChargingNeedsStatusType(str, Enum):
    """
    Accepted a SASchedule will be provided momentarily.
    Rejected Servoce is Not Available
    Processing The CSMS is gathering information to provide an SASchedule.
    """
    accepted = "Accepted"
    rejected = "Rejected"
    processing = "Processing"


class OCPPInterfaceType(str, Enum):
    """
    Enumeration of network interfaces.
    """

    wired0 = "Wired0"
    wired1 = "Wired1"
    wired2 = "Wired2"
    wired3 = "Wired3"
    wireless0 = "Wireless0"
    wireless1 = "Wireless1"
    wireless2 = "Wireless2"
    wireless3 = "Wireless3"


class OCPPTransportType(str, Enum):
    """
    Enumeration of OCPP transport mechanisms.
    SOAP is currently not a valid value for OCPP 2.0.
    """

    json = "JSON"
    soap = "SOAP"


class OCPPVersionType(str, Enum):
    """
    Enumeration of OCPP transport mechanisms.
    SOAP is currently not a valid value for OCPP 2.0.
    """

    ocpp12 = "OCPP12"
    ocpp15 = "OCPP15"
    ocpp16 = "OCPP16"
    ocpp20 = "OCPP20"


class OperationalStatusType(str, Enum):
    """
    Requested availability change in ChangeAvailability.req.
    """

    inoperative = "Inoperative"
    operative = "Operative"


class PhaseType(str, Enum):
    """
    Phase as used in SampledValue. Phase specifies how a measured value is to
    be interpreted. Please note that not all values of Phase are applicable to
    all Measurands.
    """

    l1 = "L1"
    l2 = "L2"
    l3 = "L3"
    n = "N"
    l1_n = "L1-N"
    l2_n = "L2-N"
    l3_n = "L3-N"
    l1_l2 = "L1-L2"
    l2_l3 = "L2-L3"
    l3_l1 = "L3-L1"


class PublishFirmwareStatusType(str, Enum):
    """
    Status for when publishing a Firmware
    """

    idle = "Idle"
    download_scheduled = "DownloadScheduled"
    downloading = "Downloading"
    downloaded = "Downloaded"
    published = "Published"
    download_failed = "DownloadFailed"
    download_paused = "DownloadPaused"
    invalid_checksum = "InvalidChecksum"
    checksum_verified = "ChecksumVerified"
    publish_failed = "PublishFailed"


class ReadingContextType(str, Enum):
    """
    Values of the context field of a value in SampledValue.
    """

    interruption_begin = "Interruption.Begin"
    interruption_end = "Interruption.End"
    other = "Other"
    sample_clock = "Sample.Clock"
    sample_periodic = "Sample.Periodic"
    transaction_begin = "Transaction.Begin"
    transaction_end = "Transaction.End"
    trigger = "Trigger"


class ReasonType(str, Enum):
    """
    Reason for stopping a transaction in StopTransactionRequest
    """
    de_authorized = "DeAuthorized"
    emergency_stop = "EmergencyStop"
    energy_limit_reached = "EnergyLimitReached"
    ev_disconnected = "EVDisconnected"
    ground_fault = "GroundFault"
    immediate_reset = "ImmediateReset"
    local = "Local"
    local_out_of_credit = "LocalOutOfCredit"
    master_pass = "MasterPass"
    other = "Other"
    overcurrent_fault = "OvercurrentFault"
    power_loss = "PowerLoss"
    power_quality = "PowerQuality"
    reboot = "Reboot"
    remote = "Remote"
    soc_limit_reached = "SOCLimitReached"
    stopped_by_ev = "StoppedByEV"
    time_limit_reached = "TimeLimitReached"
    timeout = "Timeout"


class RecurrencyKindType(str, Enum):
    """
    "Daily" The schedule restarts at the beginning of the next day.
    "Weekly" The schedule restarts at the beginning of the next week
              (defined as Monday morning)
    """

    daily = "Daily"
    weekly = "Weekly"


class RegistrationStatusType(str, Enum):
    """
    Result of registration in response to BootNotification.req.
    """

    accepted = "Accepted"
    pending = "Pending"
    rejected = "Rejected"


class ReportBaseType(str, Enum):
    """
    Report Base Type required in GetBaseReportRequest
    """

    configuration_inventory = "ConfigurationInventory"
    full_inventory = "FullInventory"
    summary_inventory = "SummaryInventory"


class RequestStartStopStatusType(str, Enum):
    """
    The result of a RemoteStartTransaction.req or RemoteStopTransaction.req
    request.
    """
    accepted = "Accepted"
    rejected = "Rejected"


class ReservationUpdateStatusType(str, Enum):
    expired = "Expired"
    removed = "Removed"


class ReserveNowStatusType(str, Enum):
    """
    Status in ReserveNowResponse.
    """

    accepted = "Accepted"
    faulted = "Faulted"
    occupied = "Occupied"
    rejected = "Rejected"
    unavailable = "Unavailable"


class ResetStatusType(str, Enum):
    """
    Result of Reset.req
    """

    accepted = "Accepted"
    rejected = "Rejected"
    scheduled = "Scheduled"


class ResetType(str, Enum):
    """
    Type of reset requested by Reset.req
    """
    immediate = "Immediate"
    on_idle = "OnIdle"


class SendLocalListStatusType(str, Enum):
    """
    Type of update for a SendLocalList Request.
    """

    accepted = "Accepted"
    failed = "Failed"
    version_mismatch = "VersionMismatch"


class SetMonitoringStatusType(str, Enum):
    """
    Status in SetVariableMonitoringResponse
    """

    accepted = "Accepted"
    unknown_component = "UnknownComponent"
    unknown_variable = "UnknownVariable"
    unsupported_monitor_type = "UnsupportedMonitorType"
    rejected = "Rejected"
    duplicate = "Duplicate"


class SetNetworkProfileStatusType(str, Enum):
    """
    Status in SetNetworkProfileResponse
    """

    accepted = "Accepted"
    rejected = "Rejected"
    failed = "Failed"


class SetVariableStatusType(str, Enum):
    """
    Status in ChangeConfigurationResponse.
    """

    accepted = "Accepted"
    rejected = "Rejected"
    unknown_component = "UnknownComponent"
    unknown_variable = "UnknownVariable"
    not_supported_attribute_type = "NotSupportedAttributeType"
    reboot_required = "RebootRequired"


class TransactionEventType(str, Enum):
    """
    Type of Event in TransactionEventRequest
    """

    ended = "Ended"
    started = "Started"
    updated = "Updated"


class TriggerMessageStatusType(str, Enum):
    """
    Status in TriggerMessageResponse.
    """

    accepted = "Accepted"
    rejected = "Rejected"
    not_implemented = "NotImplemented"


class TriggerReasonType(str, Enum):
    """
    Reason that triggered a transactionEventRequest
    """

    authorized = "Authorized"
    cable_plugged_in = "CablePluggedIn"
    charging_rate_changed = "ChargingRateChanged"
    charging_state_changed = "ChargingStateChanged"
    deauthorized = "Deauthorized"
    energy_limit_reached = "EnergyLimitReached"
    ev_communication_lost = "EVCommunicationLost"
    ev_connect_timeout = "EVConnectTimeout"
    meter_value_clock = "MeterValueClock"
    meter_value_periodic = "MeterValuePeriodic"
    time_limit_reached = "TimeLimitReached"
    trigger = "Trigger"
    unlock_command = "UnlockCommand"
    stop_authorized = "StopAuthorized"
    ev_departed = "EVDeparted"
    ev_detected = "EVDetected"
    remote_stop = "RemoteStop"
    remote_start = "RemoteStart"
    abnormal_condition = "AbnormalCondition"
    signed_data_received = "SignedDataReceived"
    reset_command = "ResetCommand"


class UnlockStatusType(str, Enum):
    """
    Status in response to UnlockConnector.req.
    """

    unlocked = "Unlocked"
    unlock_failed = "UnlockFailed"
    ongoing_authorized_transaction = "OngoingAuthorizedTransaction"
    unknown_connector = "UnknownConnector"


class UnpublishFirmwareStatusType(str, Enum):
    """
    Status for when unpublishing a Firmware (used by UnpublishFirmwareResponse)
    """

    download_ongoing = "DownloadOngoing"
    no_firmware = "NoFirmware"
    unpublished = "Unpublished"


class UpdateFirmwareStatusType(str, Enum):
    """
    Generic message response status for UpdateFirmwareResponse
    """

    accepted = "Accepted"
    rejected = "Rejected"
    accepted_canceled = "AcceptedCanceled"
    invalid_certificate = "InvalidCertificate"
    revoked_certificate = "RevokedCertificate"


class UpdateType(str, Enum):
    """
    Type of update for a SendLocalList Request.
    """

    differential = "Differential"
    full = "Full"


class UploadLogStatusType(str, Enum):
    """
    Status in LogStatusNotificationRequest.
    """
    bad_message = "BadMessage"
    idle = "Idle"
    not_supported_operation = "NotSupportedOperation"
    permission_denied = "PermissionDenied"
    uploaded = "Uploaded"
    upload_failure = "UploadFailure"
    uploading = "Uploading"
    accepted_canceled = "AcceptedCanceled"


class VPNType(str, Enum):
    """
    Enumeration of VPN Types used in SetNetworkProfileRequest.VPNType
    """
    ikev2 = "IKEv2"
    ipsec = "IPSec"
    l2tp = "L2TP"
    pptp = "PPTP"


# DataTypes

class UnitOfMeasureType(str, Enum):
    """
    Allowable values of the optional "unit" field of a Value element, as used
    in MeterValues.req and StopTransaction.req messages. Default value of
    "unit" is always "Wh".
    """
    asu = "ASU"
    b = "Bytes"
    db = "dB"
    dbm = "dBm"
    deg = "Deg"
    hz = "Hz"
    lx = "lx"
    m = "m"
    ms2 = "ms2"
    n = "n"
    ohm = "ohm"
    kpa = "kPa"
    percent = "Percent"
    rh = "RH"
    rpm = "RPM"
    s = "s"
    va = "VA"
    kva = "kVA"
    vah = "VAh"
    kvah = "kVAh"
    var = "var"
    kvar = "kvar"
    varh = "varh"
    kvarh = "kvarh"
    wh = "Wh"
    kwh = "kWh"
    w = "W"
    kw = "kW"
    a = "A"
    v = "V"
    celsius = "Celsius"
    fahrenheit = "Fahrenheit"
    k = "K"



from enum import Enum


class Action(str, Enum):
    """ An Action is a required part of a Call message. """
    Authorize = "Authorize"
    BootNotification = "BootNotification"
    CancelReservation = "CancelReservation"
    CertificateSigned = "CertificateSigned"
    ChangeAvailability = "ChangeAvailability"
    ChangeConfiguration = "ChangeConfiguration"
    ClearCache = "ClearCache"
    ClearChargingProfile = "ClearChargingProfile"
    DataTransfer = "DataTransfer"
    DeleteCertificate = "DeleteCertificate"
    DiagnosticsStatusNotification = "DiagnosticsStatusNotification"
    ExtendedTriggerMessage = "ExtendedTriggerMessage"
    FirmwareStatusNotification = "FirmwareStatusNotification"
    GetCompositeSchedule = "GetCompositeSchedule"
    GetConfiguration = "GetConfiguration"
    GetDiagnostics = "GetDiagnostics"
    GetInstalledCertificateIds = "GetInstalledCertificateIds"
    GetLocalListVersion = "GetLocalListVersion"
    GetLog = "GetLog"
    Heartbeat = "Heartbeat"
    InstallCertificate = "InstallCertificate"
    LogStatusNotification = "LogStatusNotification"
    MeterValues = "MeterValues"
    RemoteStartTransaction = "RemoteStartTransaction"
    RemoteStopTransaction = "RemoteStopTransaction"
    ReserveNow = "ReserveNow"
    Reset = "Reset"
    SecurityEventNotification = "SecurityEventNotification"
    SendLocalList = "SendLocalList"
    SetChargingProfile = "SetChargingProfile"
    SignCertificate = "SignCertificate"
    SignedFirmwareStatusNotification = "SignedFirmwareStatusNotification"
    SignedUpdateFirmware = "SignedUpdateFirmware"
    StartTransaction = "StartTransaction"
    StatusNotification = "StatusNotification"
    StopTransaction = "StopTransaction"
    TriggerMessage = "TriggerMessage"
    UnlockConnector = "UnlockConnector"
    UpdateFirmware = "UpdateFirmware"


class AuthorizationStatus(str, Enum):
    """
    Elements that constitute an entry of a Local Authorization List update.
    """

    accepted = "Accepted"
    blocked = "Blocked"
    expired = "Expired"
    invalid = "Invalid"
    concurrent_tx = "ConcurrentTx"


class AvailabilityStatus(str, Enum):
    """
    Status returned in response to ChangeAvailability.req.
    """

    accepted = "Accepted"
    rejected = "Rejected"
    scheduled = "Scheduled"


class AvailabilityType(str, Enum):
    """
    Requested availability change in ChangeAvailability.req.
    """

    inoperative = "Inoperative"
    operative = "Operative"


class CancelReservationStatus(str, Enum):
    """
    Status in CancelReservation.conf.
    """

    accepted = "Accepted"
    rejected = "Rejected"


class CertificateSignedStatus(str, Enum):
    """
    CertificateSignedStatusEnumType is used by: CertificateSigned.conf
    """

    accepted = "Accepted"
    rejected = "Rejected"


class CertificateStatus(str, Enum):
    """
    CertificateStatusEnumType is used by: InstallCertificate.conf
    """

    accepted = "Accepted"
    rejected = "Rejected"
    failed = "Failed"


class CertificateUse(str, Enum):
    """
    CertificateUseEnumType is used by: GetInstalledCertificateIds.req,
    InstallCertificate.req
    """

    central_system_root_certificate = "CentralSystemRootCertificate"
    manufacturer_root_certificate = "ManufacturerRootCertificate"


class ChargePointErrorCode(str, Enum):
    """
    Charge Point status reported in StatusNotification.req.
    """

    connector_lock_failure = "ConnectorLockFailure"
    ev_communication_error = "EVCommunicationError"
    ground_failure = "GroundFailure"
    high_temperature = "HighTemperature"
    internal_error = "InternalError"
    local_list_conflict = "LocalListConflict"
    no_error = "NoError"
    other_error = "OtherError"
    over_current_failure = "OverCurrentFailure"
    over_voltage = "OverVoltage"
    power_meter_failure = "PowerMeterFailure"
    power_switch_failure = "PowerSwitchFailure"
    reader_failure = "ReaderFailure"
    reset_failure = "ResetFailure"
    under_voltage = "UnderVoltage"
    weak_signal = "WeakSignal"

    # Soon to be deprecated enums
    connectorLockFailure = "ConnectorLockFailure"
    evCommunicationError = "EVCommunicationError"
    groundFailure = "GroundFailure"
    highTemperature = "HighTemperature"
    internalError = "InternalError"
    localListConflict = "LocalListConflict"
    noError = "NoError"
    otherError = "OtherError"
    overCurrentFailure = "OverCurrentFailure"
    overVoltage = "OverVoltage"
    powerMeterFailure = "PowerMeterFailure"
    powerSwitchFailure = "PowerSwitchFailure"
    readerFailure = "ReaderFailure"
    resetFailure = "ResetFailure"
    underVoltage = "UnderVoltage"
    weakSignal = "WeakSignal"


class ChargePointStatus(str, Enum):
    """
    Status reported in StatusNotification.req. A status can be reported for
    the Charge Point main controller (connectorId = 0) or for a specific
    connector. Status for the Charge Point main controller is a subset of the
    enumeration: Available, Unavailable or Faulted.

    States considered Operative are: Available, Preparing, Charging,
    SuspendedEVSE, SuspendedEV, Finishing, Reserved.
    States considered Inoperative are: Unavailable, Faulted.
    """

    available = "Available"
    preparing = "Preparing"
    charging = "Charging"
    suspended_evse = "SuspendedEVSE"
    suspended_ev = "SuspendedEV"
    finishing = "Finishing"
    reserved = "Reserved"
    unavailable = "Unavailable"
    faulted = "Faulted"

    # Soon to be deprecated enums
    suspendedevse = "SuspendedEVSE"
    suspendedev = "SuspendedEV"


class ChargingProfileKindType(str, Enum):
    """
    "Absolute": Schedule periods are relative to a fixed point in time defined
                in the schedule.
    "Recurring": Schedule restarts periodically at the first schedule period.
    "Relative": Schedule periods are relative to a situation- specific start
                point(such as the start of a session)
    """

    absolute = "Absolute"
    recurring = "Recurring"
    relative = "Relative"


class ChargingProfilePurposeType(str, Enum):
    """
    In load balancing scenarios, the Charge Point has one or more local
    charging profiles that limit the power or current to be shared by all
    connectors of the Charge Point. The Central System SHALL configure such
    a profile with ChargingProfilePurpose set to “ChargePointMaxProfile”.
    ChargePointMaxProfile can only be set at Charge Point ConnectorId 0.

    Default schedules for new transactions MAY be used to impose charging
    policies. An example could be a policy that prevents charging during
    the day. For schedules of this purpose, ChargingProfilePurpose SHALL
    be set to TxDefaultProfile. If TxDefaultProfile is set to ConnectorId 0,
    the TxDefaultProfile is applicable to all Connectors. If ConnectorId is
    set >0, it only applies to that specific connector. In the event a
    TxDefaultProfile for connector 0 is installed, and the Central
    System sends a new profile with ConnectorId >0, the TxDefaultProfile
    SHALL be replaced only for that specific connector.

    If a transaction-specific profile with purpose TxProfile is present,
    it SHALL overrule the default charging profile with purpose
    TxDefaultProfile for the duration of the current transaction only.
    After the transaction is stopped, the profile SHOULD be deleted.
    If there is no transaction active on the connector specified in a
    charging profile of type TxProfile, then the Charge Point SHALL
    discard it and return an error status in SetChargingProfile.conf.
    TxProfile SHALL only be set at Charge Point ConnectorId >0.

    It is not possible to set a ChargingProfile with purpose set to
    TxProfile without presence of an active transaction, or in advance of
    a transaction.

    In order to ensure that the updated charging profile applies only to the
    current transaction, the chargingProfilePurpose of the ChargingProfile
    MUST be set to TxProfile.
    """

    charge_point_max_profile = "ChargePointMaxProfile"
    tx_default_profile = "TxDefaultProfile"
    tx_profile = "TxProfile"

    # Soon to be deprecated enums
    chargepointmaxprofile = "ChargePointMaxProfile"
    txdefaultprofile = "TxDefaultProfile"
    txprofile = "TxProfile"


class ChargingProfileStatus(str, Enum):
    """
    Status returned in response to SetChargingProfile.req.
    """

    accepted = "Accepted"
    rejected = "Rejected"
    not_supported = "NotSupported"
    # Soon to be deprecated enums
    notSupported = "NotSupported"


class ChargingRateUnitType(str, Enum):
    """
    Unit in which a charging schedule is defined, as used in:
    GetCompositeSchedule.req and ChargingSchedule
    """

    watts = "W"
    amps = "A"


class ClearCacheStatus(str, Enum):
    """
    Status returned in response to ClearCache.req.
    """

    accepted = "Accepted"
    rejected = "Rejected"


class ClearChargingProfileStatus(str, Enum):
    """
    Status returned in response to ClearChargingProfile.req.
    """

    accepted = "Accepted"
    unknown = "Unknown"


class ConfigurationStatus(str, Enum):
    """
    Status in ChangeConfiguration.conf.
    """

    accepted = "Accepted"
    rejected = "Rejected"
    reboot_required = "RebootRequired"
    not_supported = "NotSupported"

    # Soon to be deprecated enums
    rebootRequired = "RebootRequired"
    notSupported = "NotSupported"


class DataTransferStatus(str, Enum):
    """
    Status in DataTransfer.conf.
    """
    accepted = "Accepted"
    rejected = "Rejected"
    unknown_message_id = "UnknownMessageId"
    unknown_vendor_id = "UnknownVendorId"

    # Soon to be deprecated enums
    unknownMessageId = "UnknownMessageId"
    unknownVendorId = "UnknownVendorId"


class DeleteCertificateStatus(str, Enum):
    """
    DeleteCertificateStatusEnumType is used by: DeleteCertificate.conf
    """

    accepted = "Accepted"
    failed = "Failed"
    not_found = "NotFound"


class DiagnosticsStatus(str, Enum):
    """
    Status in DiagnosticsStatusNotification.req.
    """

    idle = "Idle"
    uploaded = "Uploaded"
    upload_failed = "UploadFailed"
    uploading = "Uploading"

    # Soon to be deprecated enums
    uploadFailed = "UploadFailed"


class FirmwareStatus(str, Enum):
    """
    Status of a firmware download as reported in FirmwareStatusNotification.req
    """

    # Common for:
    # FirmwareStatusNotification.req and SignedFirmwareStatusNotification.req
    downloaded = "Downloaded"
    download_failed = "DownloadFailed"
    downloading = "Downloading"
    idle = "Idle"
    installation_failed = "InstallationFailed"
    installing = "Installing"
    installed = "Installed"

    # Only for SignedFirmwareStatusNotification.reg
    download_scheduled = "DownloadScheduled"
    download_paused = "DownloadPaused"
    install_rebooting = "InstallRebooting"
    install_scheduled = "InstallScheduled"
    install_verification_failed = "InstallVerificationFailed"
    invalid_signature = "InvalidSignature"
    signature_verified = "SignatureVerified"

    # Soon to be deprecated enums
    downloadFailed = "DownloadFailed"
    installationFailed = "InstallationFailed"


class GenericStatus(str, Enum):
    """
    Generic message response status
    """

    accepted = "Accepted"
    rejected = "Rejected"


class GetCompositeScheduleStatus(str, Enum):
    """
    Status returned in response to GetCompositeSchedule.req
    """

    accepted = "Accepted"
    rejected = "Rejected"


class GetInstalledCertificateStatus(str, Enum):
    """
    GetInstalledCertificateStatusEnumType is used by:
    GetInstalledCertificateIds.conf
    """

    accepted = "Accepted"
    not_found = "NotFound"


class HashAlgorithm(str, Enum):
    """
    HashAlgorithmEnumType is used by: CertificateHashDataType
    """

    sha256 = "SHA256"
    sha384 = "SHA384"
    sha512 = "SHA512"


class Location(str, Enum):
    """
    Allowable values of the optional "location" field of a value element in
    SampledValue.
    """

    inlet = "Inlet"
    outlet = "Outlet"
    body = "Body"
    cable = "Cable"
    ev = "EV"


class Log(str, Enum):
    """
    LogEnumType is used by GetLog.req
    """

    diagnostics_log = "DiagnosticsLog"
    security_log = "SecurityLog"


class LogStatus(str, Enum):
    """
    LogStatusEnumType is used by: GetLog.conf
    """

    accepted = "Accepted"
    rejected = "Rejected"
    accepted_canceled = "AcceptedCanceled"


class Measurand(str, Enum):
    """
    Allowable values of the optional "measurand" field of a Value element, as
    used in MeterValues.req and StopTransaction.req messages. Default value of
    "measurand" is always "Energy.Active.Import.Register"
    """

    current_export = "Current.Export"
    current_import = "Current.Import"
    current_offered = "Current.Offered"
    energy_active_export_register = "Energy.Active.Export.Register"
    energy_active_import_register = "Energy.Active.Import.Register"
    energy_reactive_export_register = "Energy.Reactive.Export.Register"
    energy_reactive_import_register = "Energy.Reactive.Import.Register"
    energy_active_export_interval = "Energy.Active.Export.Interval"
    energy_active_import_interval = "Energy.Active.Import.Interval"
    energy_reactive_export_interval = "Energy.Reactive.Export.Interval"
    energy_reactive_import_interval = "Energy.Reactive.Import.Interval"
    frequency = "Frequency"
    power_active_export = "Power.Active.Export"
    power_active_import = "Power.Active.Import"
    power_factor = "Power.Factor"
    power_offered = "Power.Offered"
    power_reactive_export = "Power.Reactive.Export"
    power_reactive_import = "Power.Reactive.Import"
    rpm = "RPM"
    soc = "SoC"
    temperature = "Temperature"
    voltage = "Voltage"

    # Soon to be deprecated enums
    currentExport = "Current.Export"
    currentImport = "Current.Import"
    currentOffered = "Current.Offered"
    energyActiveExportRegister = "Energy.Active.Export.Register"
    energyActiveImportRegister = "Energy.Active.Import.Register"
    energyReactiveExportRegister = "Energy.Reactive.Export.Register"
    energyReactiveImportRegister = "Energy.Reactive.Import.Register"
    energyActiveExportInterval = "Energy.Active.Export.Interval"
    energyActiveImportInterval = "Energy.Active.Import.Interval"
    energyReactiveExportInterval = "Energy.Reactive.Export.Interval"
    energyReactiveImportInterval = "Energy.Reactive.Import.Interval"
    powerActiveExport = "Power.Active.Export"
    powerActiveImport = "Power.Active.Import"
    powerFactor = "Power.Factor"
    powerOffered = "Power.Offered"
    powerReactiveExport = "Power.Reactive.Export"
    powerReactiveImport = "Power.Reactive.Import"


class MessageTrigger(str, Enum):
    """
    Type of request to be triggered in a TriggerMessage.req
    """

    # Common for TriggerMessage.req and ExtendedTriggerMessage.req
    boot_notification = "BootNotification"
    firmware_status_notification = "FirmwareStatusNotification"
    heartbeat = "Heartbeat"
    meter_values = "MeterValues"
    status_notification = "StatusNotification"

    # Only for TriggerMessage.req
    diagnostics_status_notification = "DiagnosticsStatusNotification"

    # Only for ExtendedTriggerMessage.req
    log_status_notification = "LogStatusNotification"
    sign_charge_point_certificate = "SignChargePointCertificate"

    # Soon to be deprecated enums
    bootNotification = "BootNotification"
    diagnosticsStatusNotification = "DiagnosticsStatusNotification"
    firmwareStatusNotification = "FirmwareStatusNotification"
    meterValues = "MeterValues"
    statusNotification = "StatusNotification"


class Phase(str, Enum):
    """
    Phase as used in SampledValue. Phase specifies how a measured value is to
    be interpreted. Please note that not all values of Phase are applicable to
    all Measurands.
    """

    l1 = "L1"
    l2 = "L2"
    l3 = "L3"
    n = "N"
    l1_n = "L1-N"
    l2_n = "L2-N"
    l3_n = "L3-N"
    l1_l2 = "L1-L2"
    l2_l3 = "L2-L3"
    l3_l1 = "L3-L1"

    # Soon to be deprecated enums
    l1n = "L1-N"
    l2n = "L2-N"
    l3n = "L3-N"
    l1l2 = "L1-L2"
    l2l3 = "L2-L3"
    l3l1 = "L3-L1"


class ReadingContext(str, Enum):
    """
    Values of the context field of a value in SampledValue.
    """

    interruption_begin = "Interruption.Begin"
    interruption_end = "Interruption.End"
    other = "Other"
    sample_clock = "Sample.Clock"
    sample_periodic = "Sample.Periodic"
    transaction_begin = "Transaction.Begin"
    transaction_end = "Transaction.End"
    trigger = "Trigger"

    # Soon to be deprecated enums
    interruptionBegin = "Interruption.Begin"
    interruptionEnd = "Interruption.End"
    sampleClock = "Sample.Clock"
    samplePeriodic = "Sample.Periodic"
    transactionBegin = "Transaction.Begin"
    transactionEnd = "Transaction.End"


class Reason(str, Enum):
    """
    Reason for stopping a transaction in StopTransaction.req.
    """

    emergency_stop = "EmergencyStop"
    ev_disconnected = "EVDisconnected"
    hard_reset = "HardReset"
    local = "Local"
    other = "Other"
    power_loss = "PowerLoss"
    reboot = "Reboot"
    remote = "Remote"
    soft_reset = "SoftReset"
    unlock_command = "UnlockCommand"
    de_authorized = "DeAuthorized"

    # Soon to be deprecated enums
    emergencyStop = "EmergencyStop"
    evDisconnected = "EVDisconnected"
    hardReset = "HardReset"
    powerLoss = "PowerLoss"
    softReset = "SoftReset"
    unlockCommand = "UnlockCommand"
    deAuthorized = "DeAuthorized"


class RecurrencyKind(str, Enum):
    """
    "Daily": The schedule restarts at the beginning of the next day.
    "Weekly": The schedule restarts at the beginning of the next week
              (defined as Monday morning)
    """

    daily = "Daily"
    weekly = "Weekly"


class RegistrationStatus(str, Enum):
    """
    Result of registration in response to BootNotification.req.
    """

    accepted = "Accepted"
    pending = "Pending"
    rejected = "Rejected"


class RemoteStartStopStatus(str, Enum):
    """
    The result of a RemoteStartTransaction.req or RemoteStopTransaction.req
    request.
    """

    accepted = "Accepted"
    rejected = "Rejected"


class ReservationStatus(str, Enum):
    """
    Status in ReserveNow.conf.
    """

    accepted = "Accepted"
    faulted = "Faulted"
    occupied = "Occupied"
    rejected = "Rejected"
    unavailable = "Unavailable"


class ResetStatus(str, Enum):
    """
    Result of Reset.req
    """

    accepted = "Accepted"
    rejected = "Rejected"


class ResetType(str, Enum):
    """
    Type of reset requested by Reset.req
    """

    hard = "Hard"
    soft = "Soft"


class TriggerMessageStatus(str, Enum):
    """
    Status in TriggerMessage.conf.
    """

    accepted = "Accepted"
    rejected = "Rejected"
    not_implemented = "NotImplemented"

    # Soon to be deprecated enums
    notImplemented = "NotImplemented"


class UnitOfMeasure(str, Enum):
    """
    Allowable values of the optional "unit" field of a Value element, as used
    in MeterValues.req and StopTransaction.req messages. Default value of
    "unit" is always "Wh".
    """

    wh = "Wh"
    kwh = "kWh"
    varh = "varh"
    kvarh = "kvarh"
    w = "W"
    kw = "kW"
    va = "VA"
    kva = "kVA"
    var = "var"
    kvar = "kvar"
    a = "A"
    v = "V"
    celsius = "Celsius"
    fahrenheit = "Fahrenheit"
    k = "K"
    percent = "Percent"
    hertz = "Hertz"


class UnlockStatus(str, Enum):
    """
    Status in response to UnlockConnector.req.
    """

    unlocked = "Unlocked"
    unlock_failed = "UnlockFailed"
    not_supported = "NotSupported"

    # Soon to be deprecated enums
    unlockFailed = "UnlockFailed"
    notSupported = "NotSupported"


class UpdateFirmwareStatus(str, Enum):
    """
    UpdateFirmwareStatusEnumType is used by: SignedUpdateFirmware.conf
    """

    accepted = "Accepted"
    rejected = "Rejected"
    accepted_canceled = "AcceptedCanceled"
    invalid_certificate = "InvalidCertificate"
    revoked_certificate = "RevokedCertificate"


class UploadLogStatus(str, Enum):
    """
    UploadLogStatusEnumType is used by: LogStatusNotification.req
    """

    bad_message = "BadMessage"
    idle = "Idle"
    not_supported_operation = "NotSupportedOperation"
    permission_denied = "PermissionDenied"
    uploaded = "Uploaded"
    upload_failure = "UploadFailure"
    uploading = "Uploading"


class UpdateStatus(str, Enum):
    """
    Type of update for a SendLocalList.req.
    """

    accepted = "Accepted"
    failed = "Failed"
    not_supported = "NotSupported"
    version_mismatch = "VersionMismatch"

    # Soon to be deprecated enums
    notSupported = "NotSupported"
    versionMismatch = "VersionMismatch"


class UpdateType(str, Enum):
    """
    Type of update for a SendLocalList.req.
    """

    differential = "Differential"
    full = "Full"


class ValueFormat(str, Enum):
    """
    Format that specifies how the value element in SampledValue is to be
    interpreted.
    """

    raw = "Raw"
    signed_data = "SignedData"

    # Soon to be deprecated enums
    signedData = "SignedData"



from typing import Any, Dict, List, Optional
from dataclasses import dataclass, field

from ocpp.v16.enums import (
    AvailabilityType,
    ChargePointErrorCode,
    ChargePointStatus,
    ChargingProfilePurposeType,
    ChargingRateUnitType,
    DiagnosticsStatus,
    FirmwareStatus,
    MessageTrigger,
    Reason,
    ResetType,
    UpdateType,
    CertificateUse,
    Log,
    UploadLogStatus,
)

# Most types of CALL messages can originate from only 1 source, either
# from a Charge Point or Central System, but not from both.
#
# Take for example the CALL for an ChangeConfiguration action. This type of
# CALL can only be send from a Central System to Charging Station, not
# the other way around.
#
# For some types of CALL messages the opposite is true; for example for the
# CALL message for a BootNotification action. This can only come from a Charge
# Point and send to a Central System.
#
# The only CALL that can originate from both a Central System and a
# Charge Point is the CALL message for a DataTransfer.

# The now following section of classes are for CALL messages that flow
# from Central System to Charge Point.


@dataclass
class CancelReservationPayload:
    reservation_id: int


@dataclass
class CertificateSignedPayload:
    certificate_chain: str


@dataclass
class ChangeAvailabilityPayload:
    connector_id: int
    type: AvailabilityType


@dataclass
class ChangeConfigurationPayload:
    key: str
    value: Any


@dataclass
class ClearCachePayload:
    pass


@dataclass
class ClearChargingProfilePayload:
    id: Optional[int] = None
    connector_id: Optional[int] = None
    charging_profile_purpose: Optional[ChargingProfilePurposeType] = None
    stack_level: Optional[int] = None


@dataclass
class DeleteCertificatePayload:
    certificate_hash_data: Dict


@dataclass
class ExtendedTriggerMessagePayload:
    requested_message: MessageTrigger
    connector_id: Optional[int] = None


@dataclass
class GetCompositeSchedulePayload:
    connector_id: int
    duration: int
    charging_rate_unit: Optional[ChargingRateUnitType] = None


@dataclass
class GetConfigurationPayload:
    key: Optional[List] = None


@dataclass
class GetDiagnosticsPayload:
    location: str
    retries: Optional[int] = None
    retry_interval: Optional[int] = None
    start_time: Optional[str] = None
    stop_time: Optional[str] = None


@dataclass
class GetInstalledCertificateIdsPayload:
    certificate_type: CertificateUse


@dataclass
class GetLocalListVersionPayload:
    pass


@dataclass
class GetLogPayload:
    log: Dict
    log_type: Log
    request_id: int
    retries: Optional[int] = None
    retry_interval: Optional[int] = None


@dataclass
class InstallCertificatePayload:
    certificate_type: CertificateUse
    certificate: str


@dataclass
class RemoteStartTransactionPayload:
    id_tag: str
    connector_id: Optional[int] = None
    charging_profile: Optional[Dict] = None


@dataclass
class RemoteStopTransactionPayload:
    transaction_id: int


@dataclass
class ReserveNowPayload:
    connector_id: int
    expiry_date: str
    id_tag: str
    reservation_id: int
    parent_id_tag: Optional[str] = None


@dataclass
class ResetPayload:
    type: ResetType


@dataclass
class SendLocalListPayload:
    list_version: int
    update_type: UpdateType
    local_authorization_list: List = field(default_factory=list)


@dataclass
class SetChargingProfilePayload:
    connector_id: int
    cs_charging_profiles: Dict


@dataclass
class SignedUpdateFirmwarePayload:
    request_id: int
    firmware: Dict
    retries: Optional[int] = None
    retry_interval: Optional[int] = None


@dataclass
class TriggerMessagePayload:
    requested_message: MessageTrigger
    connector_id: Optional[int] = None


@dataclass
class UnlockConnectorPayload:
    connector_id: int


@dataclass
class UpdateFirmwarePayload:
    location: str
    retrieve_date: str
    retries: Optional[int] = None
    retry_interval: Optional[int] = None


# The CALL messages that flow from Charge Point to Central System are listed
# in the bottom part of this module.

@dataclass
class AuthorizePayload:
    id_tag: str


@dataclass
class BootNotificationPayload:
    charge_point_model: str
    charge_point_vendor: str
    charge_box_serial_number: Optional[str] = None
    charge_point_serial_number: Optional[str] = None
    firmware_version: Optional[str] = None
    iccid: Optional[str] = None
    imsi: Optional[str] = None
    meter_serial_number: Optional[str] = None
    meter_type: Optional[str] = None


@dataclass
class DiagnosticsStatusNotificationPayload:
    status: DiagnosticsStatus


@dataclass
class FirmwareStatusNotificationPayload:
    status: FirmwareStatus


@dataclass
class HeartbeatPayload:
    pass


@dataclass
class LogStatusNotificationPayload:
    status: UploadLogStatus
    request_id: int


@dataclass
class MeterValuesPayload:
    connector_id: int
    meter_value: List = field(default_factory=list)
    transaction_id: Optional[int] = None


@dataclass
class SecurityEventNotificationPayload:
    type: str
    timestamp: str
    tech_info: Optional[str]


@dataclass
class SignCertificatePayload:
    csr: str


@dataclass
class SignedFirmwareStatusNotificationPayload:
    status: FirmwareStatus
    request_id: int


@dataclass
class StartTransactionPayload:
    connector_id: int
    id_tag: str
    meter_start: int
    timestamp: str
    reservation_id: Optional[int] = None


@dataclass
class StopTransactionPayload:
    meter_stop: int
    timestamp: str
    transaction_id: int
    reason: Optional[Reason] = None
    id_tag: Optional[str] = None
    transaction_data: Optional[List] = None


@dataclass
class StatusNotificationPayload:
    connector_id: int
    error_code: ChargePointErrorCode
    status: ChargePointStatus
    timestamp: Optional[str] = None
    info: Optional[str] = None
    vendor_id: Optional[str] = None
    vendor_error_code: Optional[str] = None


# The DataTransfer CALL can be send both from Central System as well as from a
# Charge Point.


@dataclass
class DataTransferPayload:
    vendor_id: str
    message_id: Optional[str] = None
    data: Optional[str] = None



""" Module containing classes that model the several OCPP messages types. It
also contain some helper functions for packing and unpacking messages.  """
import os
import json
import decimal
from typing import Callable, Dict
from dataclasses import asdict, is_dataclass

from jsonschema import Draft4Validator
from jsonschema.exceptions import ValidationError as SchemaValidationError

from ocpp.exceptions import (OCPPError, FormatViolationError,
                             PropertyConstraintViolationError,
                             ProtocolError, ValidationError,
                             UnknownCallErrorCodeError)

_validators: Dict[str, Draft4Validator] = {}


class _DecimalEncoder(json.JSONEncoder):
    """ Encode values of type `decimal.Decimal` using 1 decimal point.

    A custom encoder is required because `json.dumps()` cannot encode a value
    of type decimal.Decimal. This raises a TypeError:

        >>> import decimal
        >>> import json
        >>> >>> json.dumps(decimal.Decimal(3))
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "/home/developer/.pyenv/versions/3.7.0/lib/python3.7/json/__init__.py", line 231, in dumps  # noqa
            return _default_encoder.encode(obj)
          File "/home/developer/.pyenv/versions/3.7.0/lib/python3.7/json/encoder.py", line 199, in encode
            chunks = self.iterencode(o, _one_shot=True)
          File "/home/developer/.pyenv/versions/3.7.0/lib/python3.7/json/encoder.py", line 257, in iterencode
            return _iterencode(o, 0)
          File "/home/developer/.pyenv/versions/3.7.0/lib/python3.7/json/encoder.py", line 179, in default
            raise TypeError(f'Object of type {o.__class__.__name__} '
        TypeError: Object of type Decimal is not JSON serializable

    This can be prevented by using a custom encoder.

    """
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float("%.1f" % obj)
        return json.JSONEncoder.default(self, obj)


class MessageType:
    """ Number identifying the different types of OCPP messages. """
    #: Call identifies a request.
    Call = 2

    #: CallResult identifies a successful response.
    CallResult = 3

    #: CallError identifies an erroneous response.
    CallError = 4


def unpack(msg):
    """
    Unpacks a message into either a Call, CallError or CallResult.
    """
    try:
        msg = json.loads(msg)
    except json.JSONDecodeError as e:
        raise FormatViolationError(f'Message is not valid JSON: {e}')

    if not isinstance(msg, list):
        raise ProtocolError("OCPP message hasn't the correct format. It "
                            f"should be a list, but got {type(msg)} instead")

    for cls in [Call, CallResult, CallError]:
        try:
            if msg[0] == cls.message_type_id:
                return cls(*msg[1:])
        except IndexError:
            raise ProtocolError("Message doesn\'t contain MessageTypeID")

    raise PropertyConstraintViolationError(f"MessageTypeId '{msg[0]}' isn't "
                                           "valid")


def pack(msg):
    """
    Returns the JSON representation of a Call, CallError or CallResult.

    It just calls the 'to_json()' method of the message. But it is here mainly
    to complement the 'unpack' function of this module.
    """
    return msg.to_json()


def get_validator(
        message_type_id: int,
        action: str,
        ocpp_version: str,
        parse_float: Callable = float
) -> Draft4Validator:
    """
    Read schema from disk and return as `Draft4Validator`. Instances will be
    cached for performance reasons.

    The `parse_float` argument can be used to set the conversion method that
    is used to parse floats. It must be a callable taking 1 argument. By
    default it is `float()`, but certain schema's require `decimal.Decimal()`.
    """
    if ocpp_version not in ["1.6", "2.0", "2.0.1"]:
        raise ValueError

    schemas_dir = 'v' + ocpp_version.replace('.', '')

    schema_name = action
    if message_type_id == MessageType.CallResult:
        schema_name += 'Response'
    elif message_type_id == MessageType.Call:
        if ocpp_version in ["2.0", "2.0.1"]:
            schema_name += 'Request'

    if ocpp_version == "2.0":
        schema_name += '_v1p0'

    cache_key = schema_name + '_' + ocpp_version
    if cache_key in _validators:
        return _validators[cache_key]

    dir,  _ = os.path.split(os.path.realpath(__file__))
    relative_path = f'{schemas_dir}/schemas/{schema_name}.json'
    path = os.path.join(dir, relative_path)

    # The JSON schemas for OCPP 2.0 start with a byte order mark (BOM)
    # character. If no encoding is given, reading the schema would fail with:
    #
    #     Unexpected UTF-8 BOM (decode using utf-8-sig):
    with open(path, 'r', encoding='utf-8-sig') as f:
        data = f.read()
        validator = Draft4Validator(json.loads(data, parse_float=parse_float))
        _validators[cache_key] = validator

    return _validators[cache_key]


def validate_payload(message, ocpp_version):
    """ Validate the payload of the message using JSON schemas. """
    if type(message) not in [Call, CallResult]:
        raise ValidationError("Payload can't be validated because message "
                              f"type. It's '{type(message)}', but it should "
                              "be either 'Call'  or 'CallResult'.")

    try:
        # 3 OCPP 1.6 schedules have fields of type floats. The JSON schema
        # defines a certain precision for these fields of 1 decimal. A value of
        # 21.4 is valid, whereas a value if 4.11 is not.
        #
        # The problem is that Python's internal representation of 21.4 might
        # have more than 1 decimal. It might be 21.399999999999995. This would
        # make the validation fail, although the payload is correct. This is a
        # known issue with jsonschemas, see:
        # https://github.com/Julian/jsonschema/issues/247
        #
        # This issue can be fixed by using a different parser for floats than
        # the default one that is used.
        #
        # Both the schema and the payload must be parsed using the different
        # parser for floats.
        if ocpp_version == '1.6' and (
            (type(message) == Call and
                message.action in ['SetChargingProfile', 'RemoteStartTransaction'])  # noqa
            or
            (type(message) == CallResult and
                message.action == 'GetCompositeSchedule')
        ):
            validator = get_validator(
                message.message_type_id, message.action,
                ocpp_version, parse_float=decimal.Decimal
            )

            message.payload = json.loads(
                json.dumps(message.payload), parse_float=decimal.Decimal
            )
        else:
            validator = get_validator(
                message.message_type_id, message.action, ocpp_version
            )
    except (OSError, json.JSONDecodeError) as e:
        raise ValidationError("Failed to load validation schema for action "
                              f"'{message.action}': {e}")

    try:
        validator.validate(message.payload)
    except SchemaValidationError as e:
        raise ValidationError(f"Payload '{message.payload} for action "
                              f"'{message.action}' is not valid: {e}")


class Call:
    """ A Call is a type of message that initiate a request/response sequence.
    Both central systems and charge points can send this message.

    From the specification:

        A Call always consists of 4 elements: The standard elements
        MessageTypeId and UniqueId, a specific Action that is required on the
        other side and a payload, the arguments to the Action. The syntax of a
        call looks like this:

            [<MessageTypeId>, "<UniqueId>", "<Action>", {<Payload>}]

        ...

        For example, a BootNotification request could look like this:

            [2,
             "19223201",
             "BootNotification",
             {
              "chargePointVendor": "VendorX",
              "chargePointModel": "SingleSocketCharger"
             }
            ]
    """
    message_type_id = 2

    def __init__(self, unique_id, action, payload):
        self.unique_id = unique_id
        self.action = action
        self.payload = payload

        if is_dataclass(payload):
            self.payload = asdict(payload)

    def to_json(self):
        """ Return a valid JSON representation of the instance. """
        return json.dumps([
            self.message_type_id,
            self.unique_id,
            self.action,
            self.payload,
        ],
            # By default json.dumps() adds a white space after every separator.
            # By setting the separator manually that can be avoided.
            separators=(',', ':'),
            cls=_DecimalEncoder
        )

    def create_call_result(self, payload):
        call_result = CallResult(self.unique_id, payload)
        call_result.action = self.action
        return call_result

    def create_call_error(self, exception):
        error_code = "InternalError"
        error_description = "An unexpected error occurred."
        error_details = {}

        if isinstance(exception, OCPPError):
            error_code = exception.code
            error_description = exception.description
            error_details = exception.details

        return CallError(
            self.unique_id,
            error_code,
            error_description,
            error_details,
        )

    def __repr__(self):
        return f"<Call - unique_id={self.unique_id}, action={self.action}, " \
               f"payload={self.payload}>"


class CallResult:
    """
    A CallResult is a message indicating that a Call has been handled
    succesfully.

    From the specification:

        A CallResult always consists of 3 elements: The standard elements
        MessageTypeId, UniqueId and a payload, containing the response to the
        Action in the original Call. The syntax of a call looks like this:

            [<MessageTypeId>, "<UniqueId>", {<Payload>}]

        ...

        For example, a BootNotification response could look like this:

            [3,
             "19223201",
             {
              "status":"Accepted",
              "currentTime":"2013-02-01T20:53:32.486Z",
              "heartbeatInterval":300
             }
            ]

    """
    message_type_id = 3

    def __init__(self, unique_id, payload, action=None):
        self.unique_id = unique_id
        self.payload = payload

        # Strictly speaking no action is required in a CallResult. But in order
        # to validate the message it is needed.
        self.action = action

    def to_json(self):
        return json.dumps([
            self.message_type_id,
            self.unique_id,
            self.payload,
        ],
            # By default json.dumps() adds a white space after every separator.
            # By setting the separator manually that can be avoided.
            separators=(',', ':'),
            cls=_DecimalEncoder
        )

    def __repr__(self):
        return f"<CallResult - unique_id={self.unique_id}, " \
               f"action={self.action}, " \
               f"payload={self.payload}>"


class CallError:
    """
    A CallError is a response to a Call that indicates an error.

    From the specification:

        CallError always consists of 5 elements: The standard elements
        MessageTypeId and UniqueId, an errorCode string, an errorDescription
        string and an errorDetails object.

        The syntax of a call looks like this:

            [<MessageTypeId>, "<UniqueId>", "<errorCode>", "<errorDescription>", {<errorDetails>}] # noqa
    """
    message_type_id = 4

    def __init__(self, unique_id, error_code, error_description,
                 error_details=None):
        self.unique_id = unique_id
        self.error_code = error_code
        self.error_description = error_description
        self.error_details = error_details

    def to_json(self):
        return json.dumps([
            self.message_type_id,
            self.unique_id,
            self.error_code,
            self.error_description,
            self.error_details,
        ],
            # By default json.dumps() adds a white space after every separator.
            # By setting the separator manually that can be avoided.
            separators=(',', ':'),
            cls=_DecimalEncoder
        )

    def to_exception(self):
        """ Return the exception that corresponds to the CallError. """
        for error in OCPPError.__subclasses__():
            if error.code == self.error_code:
                return error(
                    description=self.error_description,
                    details=self.error_details
                )

        raise UnknownCallErrorCodeError("Error code '%s' is not defined by the"
                                        " OCPP specification", self.error_code)

    def __repr__(self):
        return f"<CallError - unique_id={self.unique_id}, " \
               f"error_code={self.error_code}, " \
               f"error_description={self.error_description}, " \
               f"error_details={self.error_details}>"



from typing import Dict, List, Optional
from dataclasses import dataclass


@dataclass
class AuthorizePayload:
    id_token: Dict
    certificate: Optional[str] = None
    iso15118_certificate_hash_data: Optional[List] = None


@dataclass
class BootNotificationPayload:
    charging_station: Dict
    reason: str


@dataclass
class CancelReservationPayload:
    reservation_id: int


@dataclass
class CertificateSignedPayload:
    certificate_chain: str
    certificate_type: Optional[str] = None


@dataclass
class ChangeAvailabilityPayload:
    operational_status: str
    evse: Optional[Dict] = None


@dataclass
class ClearCachePayload:
    pass


@dataclass
class ClearChargingProfilePayload:
    charging_profile_id: Optional[int] = None
    charging_profile_criteria: Optional[Dict] = None


@dataclass
class ClearDisplayMessagePayload:
    id: int


@dataclass
class ClearVariableMonitoringPayload:
    id: List


@dataclass
class ClearedChargingLimitPayload:
    charging_limit_source: str
    evse_id: Optional[int] = None


@dataclass
class CostUpdatedPayload:
    total_cost: int
    transaction_id: str


@dataclass
class CustomerInformationPayload:
    request_id: int
    report: bool
    clear: bool
    customer_certificate: Optional[Dict] = None
    id_token: Optional[Dict] = None
    customer_identifier: Optional[str] = None


@dataclass
class DataTransferPayload:
    vendor_id: str
    message_id: Optional[str] = None
    data: Optional[str] = None


@dataclass
class DeleteCertificatePayload:
    certificate_hash_data: Dict


@dataclass
class FirmwareStatusNotificationPayload:
    status: str
    request_id: Optional[int] = None


@dataclass
class Get15118EVCertificatePayload:
    iso15118_schema_version: str
    action: str
    exi_request: str


@dataclass
class GetBaseReportPayload:
    request_id: int
    report_base: str


@dataclass
class GetCertificateStatusPayload:
    ocsp_request_data: Dict


@dataclass
class GetChargingProfilesPayload:
    request_id: int
    charging_profile: Dict
    evse_id: Optional[int] = None


@dataclass
class GetCompositeSchedulePayload:
    duration: int
    evse_id: int
    charging_rate_unit: Optional[str] = None


@dataclass
class GetDisplayMessagesPayload:
    request_id: int
    id: Optional[List] = None
    priority: Optional[str] = None
    state: Optional[str] = None


@dataclass
class GetInstalledCertificateIdsPayload:
    certificate_type: Optional[List] = None


@dataclass
class GetLocalListVersionPayload:
    pass


@dataclass
class GetLogPayload:
    log: Dict
    log_type: str
    request_id: int
    retries: Optional[int] = None
    retry_interval: Optional[int] = None


@dataclass
class GetMonitoringReportPayload:
    request_id: int
    component_variable: Optional[List] = None
    monitoring_criteria: Optional[List] = None


@dataclass
class GetReportPayload:
    request_id: int
    component_variable: Optional[List] = None
    component_criteria: Optional[List] = None


@dataclass
class GetTransactionStatusPayload:
    transaction_id: Optional[str] = None


@dataclass
class GetVariablesPayload:
    get_variable_data: List


@dataclass
class HeartbeatPayload:
    pass


@dataclass
class InstallCertificatePayload:
    certificate_type: str
    certificate: str


@dataclass
class LogStatusNotificationPayload:
    status: str
    request_id: Optional[int] = None


@dataclass
class MeterValuesPayload:
    evse_id: int
    meter_value: List


@dataclass
class NotifyChargingLimitPayload:
    charging_limit: Dict
    charging_schedule: Optional[List] = None
    evse_id: Optional[int] = None


@dataclass
class NotifyCustomerInformationPayload:
    data: str
    seq_no: int
    generated_at: str
    request_id: int
    tbc: Optional[bool] = None


@dataclass
class NotifyDisplayMessagesPayload:
    request_id: int
    message_info: Optional[List] = None
    tbc: Optional[bool] = None


@dataclass
class NotifyEVChargingNeedsPayload:
    charging_needs: Dict
    evse_id: int
    max_schedule_tuples: Optional[int] = None


@dataclass
class NotifyEVChargingSchedulePayload:
    time_base: str
    charging_schedule: Dict
    evse_id: int


@dataclass
class NotifyEventPayload:
    generated_at: str
    seq_no: int
    event_data: List
    tbc: Optional[bool] = None


@dataclass
class NotifyMonitoringReportPayload:
    request_id: int
    seq_no: int
    generated_at: str
    monitor: Optional[List] = None
    tbc: Optional[bool] = None


@dataclass
class NotifyReportPayload:
    request_id: int
    generated_at: str
    seq_no: int
    report_data: Optional[List] = None
    tbc: Optional[bool] = None


@dataclass
class PublishFirmwarePayload:
    location: str
    checksum: str
    request_id: int
    retries: Optional[int] = None
    retry_interval: Optional[int] = None


@dataclass
class PublishFirmwareStatusNotificationPayload:
    status: str
    location: Optional[List] = None
    request_id: Optional[int] = None


@dataclass
class ReportChargingProfilesPayload:
    request_id: int
    charging_limit_source: str
    charging_profile: List
    evse_id: int
    tbc: Optional[bool] = None


@dataclass
class RequestStartTransactionPayload:
    id_token: Dict
    remote_start_id: int
    evse_id: Optional[int] = None
    group_id_token: Optional[Dict] = None
    charging_profile: Optional[Dict] = None


@dataclass
class RequestStopTransactionPayload:
    transaction_id: str


@dataclass
class ReservationStatusUpdatePayload:
    reservation_id: int
    reservation_update_status: str


@dataclass
class ReserveNowPayload:
    id: int
    expiry_date_time: str
    id_token: Dict
    connector_type: Optional[str] = None
    evse_id: Optional[int] = None
    group_id_token: Optional[Dict] = None


@dataclass
class ResetPayload:
    type: str
    evse_id: Optional[int] = None


@dataclass
class SecurityEventNotificationPayload:
    type: str
    timestamp: str
    tech_info: Optional[str] = None


@dataclass
class SendLocalListPayload:
    version_number: int
    update_type: str
    local_authorization_list: Optional[List] = None


@dataclass
class SetChargingProfilePayload:
    evse_id: int
    charging_profile: Dict


@dataclass
class SetDisplayMessagePayload:
    message: Dict


@dataclass
class SetMonitoringBasePayload:
    monitoring_base: str


@dataclass
class SetMonitoringLevelPayload:
    severity: int


@dataclass
class SetNetworkProfilePayload:
    configuration_slot: int
    connection_data: Dict


@dataclass
class SetVariableMonitoringPayload:
    set_monitoring_data: List


@dataclass
class SetVariablesPayload:
    set_variable_data: List


@dataclass
class SignCertificatePayload:
    csr: str
    certificate_type: Optional[str] = None


@dataclass
class StatusNotificationPayload:
    timestamp: str
    connector_status: str
    evse_id: int
    connector_id: int


@dataclass
class TransactionEventPayload:
    event_type: str
    timestamp: str
    trigger_reason: str
    seq_no: int
    transaction_info: Dict
    meter_value: Optional[List] = None
    offline: Optional[bool] = None
    number_of_phases_used: Optional[int] = None
    cable_max_current: Optional[int] = None
    reservation_id: Optional[int] = None
    evse: Optional[Dict] = None
    id_token: Optional[Dict] = None


@dataclass
class TriggerMessagePayload:
    requested_message: str
    evse: Optional[Dict] = None


@dataclass
class UnlockConnectorPayload:
    evse_id: int
    connector_id: int


@dataclass
class UnpublishFirmwarePayload:
    checksum: str


@dataclass
class UpdateFirmwarePayload:
    request_id: int
    firmware: Dict
    retries: Optional[int] = None
    retry_interval: Optional[int] = None



import functools

routables = []


def on(action, *, skip_schema_validation=False):
    """
    Function decorator to mark function as handler for specific action. The
    wrapped function may be async or sync.

    The handler function will receive keyword arguments derived from the
    payload of the specific action. It's recommended you use `**kwargs` in your
    definition to ignore any extra arguments that may be added in the future.

    The handler function should return a relevant payload to be returned to the
    Charge Point.

    It can be used like so:

    ```
    class MyChargePoint(cp):
        @on(Action.BootNotification):
        async def on_boot_notification(
            self,
            charge_point_model,
            charge_point_vendor,
            **kwargs,
        ):
            print(f'{charge_point_model} from {charge_point_vendor} booted.')

            now = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S') + "Z"
            return call_result.BootNotificationPayload(
                current_time=now,
                interval=30,
                status="Accepted",
            )
    ```

    The decorator takes an optional argument `skip_schema_validation` which
    defaults to False. Setting this argument to `True` will disable schema
    validation of the request and the response of the specific route.

    """
    def decorator(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        inner._on_action = action
        inner._skip_schema_validation = skip_schema_validation
        if func.__name__ not in routables:
            routables.append(func.__name__)
        return inner

    return decorator


def after(action):
    """ Function decorator to mark function as hook to post-request hook.

    This hook's arguments are the data that is in the payload for the specific
    action.

    It can be used like so:

        @after(Action.BootNotification):
        def after_boot_notification():
            pass

    """
    def decorator(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        inner._after_action = action
        if func.__name__ not in routables:
            routables.append(func.__name__)
        return inner
    return decorator


def create_route_map(obj):
    """
    Iterates of all attributes of the class looking for attributes which
    have been decorated by the @on() decorator It returns a dictionary where
    the action name are the keys and the decorated functions are the values.

    To illustrate this with an example, consider the following function:

        class ChargePoint:

            @on(Action.BootNotification)
            def on_boot_notification(self, *args, **kwargs):
                pass

            @after(Action.BootNotification)
            def after_boot_notification(self, *args, **kwargs):
                pass


    In this case this returns:

        {
            Action.BootNotification: {
                '_on_action': <reference to 'on_boot_notification'>,
                '_after_action': <reference to 'after_boot_notification'>,
                '_skip_schema_validation': False,
            },
        }

    """
    routes = {}
    for attr_name in routables:
        for option in ['_on_action', '_after_action']:
            try:
                attr = getattr(obj, attr_name)
                action = getattr(attr, option)

                if action not in routes:
                    routes[action] = {}

                # Routes decorated with the `@on()` decorator can be configured
                # to skip validation of the input and output. For more info see
                # the docstring of `on()`.
                if option == '_on_action':
                    routes[action]['_skip_schema_validation'] = \
                        getattr(attr, '_skip_schema_validation', False)

                routes[action][option] = attr

            except AttributeError:
                continue

    return routes



from dataclasses import dataclass
from typing import Optional

from ocpp.v201 import enums


@dataclass
class ACChargingParametersType:
    """
    EV AC charging parameters.
    ACChargingParametersType is used by: ChargingNeedsType
    """

    energy_amount: int
    ev_min_current: int
    ev_max_current: int
    ev_max_voltage: int


@dataclass
class AdditionalInfoType:
    """
    Contains a case insensitive identifier to use for the authorization and the
    type of authorization to support multiple forms of identifiers.
    AdditionalInfoType is used by: IdTokenType
    """

    additional_id_token: str
    type: str


@dataclass
class APNType:
    """
    Collection of configuration data needed to make a data-connection over a
    cellular network.
    APNType is used by: SetNetworkProfileRequest.NetworkConnectionProfileType
    """

    apn: str
    apn_authentication: enums.APNAuthenticationType
    apn_user_name: Optional[str] = None
    apn_password: Optional[str] = None
    sim_pin: Optional[int] = None
    peferred_network: Optional[str] = None
    use_only_preferred_network: Optional[bool] = None


@dataclass
class CertificateHashDataType:
    """
    CertificateHashDataType is used by: CertificateHashDataChainType,
    DeleteCertificateRequest, CustomerInformationRequest
    """

    hash_algorithm: enums.HashAlgorithmType
    issuer_name_hash: str
    issuer_key_hash: str
    serial_number: str


@dataclass
class CertificateHashDataChainType:
    """
    CertificateHashDataChainType is used by: GetInstalledCertificateIdsResponse
    """

    certificate_type: enums.GetCertificateIdUseType
    certificate_hash_data: CertificateHashDataType
    child_certificate_hash_data: CertificateHashDataType


@dataclass
class ChargingLimitType:
    """ChargingLimitType is used by: NotifyChargingLimitRequest"""

    charging_limit_source: enums.ChargingLimitSourceType
    is_grid_critical: Optional[bool] = None


@dataclass
class DCChargingParametersType:
    """
    EV DC charging parameters
    DCChargingParametersType is used by: ChargingNeedsType
    """

    ev_max_current: int
    ev_max_voltage: int
    energy_amount: Optional[int] = None
    ev_max_power: Optional[int] = None
    state_of_charge: Optional[int] = None
    ev_energy_capacity: Optional[int] = None
    full_soc: Optional[int] = None
    bulk_soc: Optional[int] = None


@dataclass
class ChargingNeedsType:
    """
    ChargingNeedsType is used by: NotifyEVChargingNeedsRequest
    """

    request_energy_transfer: enums.EnergyTransferModeType
    departure_time: Optional[str] = None
    ac_charging_parameters: Optional[ACChargingParametersType] = None
    dc_charging_parameters: Optional[DCChargingParametersType] = None


@dataclass
class ChargingProfileCriterionType:
    """
    A ChargingProfile consists of ChargingSchedule, describing the amount of
    power or current that can be delivered per time interval.
    ChargingProfileCriterionType is used by: GetChargingProfilesRequest
    """

    charging_profile_purpose: Optional[enums.ChargingProfilePurposeType] = None
    stack_level: Optional[int] = None
    charging_profile_id: Optional[int] = None
    charging_limit_source: Optional[enums.ChargingLimitSourceType] = None


@dataclass
class ChargingSchedulePeriodType:
    """
    Charging schedule period structure defines a time period in a charging
    schedule.
    ChargingSchedulePeriodType is used by: ChargingScheduleType,
    CompositeScheduleType
    """

    start_period: int
    limit: float
    number_phases: Optional[int] = None
    phase_to_use: Optional[int] = None


@dataclass
class RelativeTimeIntervalType:
    """RelativeTimeIntervalType is used by: SalesTariffEntryType"""

    start: int
    duration: Optional[int] = None


@dataclass
class CostType:
    """
    CostType is used by: ConsumptionCostType
    """

    cost_kind: enums.CostKindType
    amount: int
    amount_multiplier: Optional[int] = None


@dataclass
class ConsumptionCostType:
    """
    ConsumptionCostType is used by: SalesTariffEntryType
    """

    start_value: float
    cost: CostType


@dataclass
class SalesTariffEntryType:
    """SalesTariffEntryType is used by: SalesTariffType"""

    relative_time_interval: RelativeTimeIntervalType
    consumption_cost: Optional[ConsumptionCostType] = None
    e_price_level: Optional[int] = None


@dataclass
class SalesTariffType:
    """
    This dataType is based on dataTypes from ISO 15118-2.
    SalesTariffType is used by: ChargingScheduleType
    """

    id: int
    sales_tariff_entry: SalesTariffEntryType
    sales_tariff_description: Optional[str] = None
    num_e_price_levels: Optional[int] = None


@dataclass
class ChargingScheduleType:
    """
    Charging schedule structure defines a list of charging periods, as used in:
    GetCompositeSchedule.conf and ChargingProfile.
    ChargingScheduleType is used by: ChargingProfileType,
    NotifyChargingLimitRequest, NotifyEVChargingScheduleRequest
    """

    id: int
    charging_rate_unit: enums.ChargingRateUnitType
    charging_schedule_period: ChargingSchedulePeriodType
    start_schedule: Optional[str] = None
    duration: Optional[int] = None
    min_charging_rate: Optional[float] = None
    sales_tariff: Optional[SalesTariffType] = None


@dataclass
class ChargingProfileType:
    """
    A ChargingProfile consists of ChargingSchedule, describing the amount of
    power or current that can be delivered per time interval.
    ChargingProfileType is used by: RequestStartTransactionRequest,
    SetChargingProfileRequest, ReportChargingProfilesRequest
    """

    id: int
    stack_level: int
    charging_profile_purpose: enums.ChargingProfilePurposeType
    charging_profile_kind: enums.ChargingProfileKindType
    charging_schedule: ChargingScheduleType
    valid_from: Optional[str] = None
    valid_to: Optional[str] = None
    transaction_id: Optional[str] = None
    recurrency_kind: Optional[enums.RecurrencyKindType] = None


@dataclass
class ClearChargingProfileType:
    """
    A ChargingProfile consists of a ChargingSchedule, describing the amount of
    power or current that can be delivered per time interval.
    ClearChargingProfileType is used by: ClearChargingProfileRequest
    """

    evse_id: Optional[int] = None
    charging_profile_purpose: Optional[enums.ChargingProfilePurposeType] = None
    stack_level: Optional[int] = None


@dataclass
class StatusInfoType:
    """
    Element providing more information about the status.
    StatusInfoType is used by: ClearMonitoringResultType,
    BootNotificationResponse, CancelReservationResponse,
    TriggerMessageResponse, UnlockConnectorResponse, UpdateFirmwareResponse,
    ClearDisplayMessageResponse, Get15118EVCertificateResponse,
    GetCompositeScheduleResponse, ChangeAvailabilityResponse, GetLogResponse,
    ClearChargingProfileResponse, NotifyEVChargingNeedsResponse,
    ClearCacheResponse, NotifyEVChargingScheduleResponse,
    RequestStartTransactionResponse, RequestStopTransactionResponse,
    SetChargingProfileResponse, SetDisplayMessageResponse,
    SetNetworkProfileResponse, SignCertificateResponse, DataTransferResponse,
    CertificateSignedResponse, DeleteCertificateResponse,
    GetChargingProfilesResponse, GetInstalledCertificateIdsResponse,
    InstallCertificateResponse, GetBaseReportResponse,
    GetMonitoringReportResponse, GetReportResponse,
    GetVariablesResponse.GetVariableResultType, ReserveNowResponse,
    SetMonitoringBaseResponse, SetMonitoringLevelResponse,
    SetVariableMonitoringResponse.SetMonitoringResultType,
    SetVariablesResponse.SetVariableResultType, PublishFirmwareResponse,
    GetCertificateStatusResponse, ResetResponse, GetDisplayMessagesResponse,
    CustomerInformationResponse, SendLocalListResponse
    """

    reason_code: str
    additional_info: Optional[str] = None


@dataclass
class ClearMonitoringResultType:
    """
    ClearMonitoringResultType is used by: ClearVariableMonitoringResponse
    """

    status: enums.ClearMonitoringStatusType
    id: int
    status_info: Optional[StatusInfoType] = None


@dataclass
class EVSEType:
    """
    Electric Vehicle Supply Equipment
    EVSEType is used by: ComponentType, TriggerMessageRequest,
    ChangeAvailabilityRequest, TransactionEventRequest
    """

    id: int
    connector_id: Optional[int] = None


@dataclass
class ComponentType:
    """
    A physical or logical component.
    ComponentType is used by: ComponentVariableType, MessageInfoType,
    GetVariablesRequest.GetVariableDataType,
    GetVariablesResponse.GetVariableResultType,
    NotifyMonitoringReportRequest.MonitoringDataType,
    NotifyReportRequest.ReportDataType,
    SetVariableMonitoringRequest.SetMonitoringDataType,
    SetVariableMonitoringResponse.SetMonitoringResultType,
    SetVariablesRequest.SetVariableDataType,
    SetVariablesResponse.SetVariableResultType,
    NotifyEventRequest.EventDataType
    """

    name: str
    instance: Optional[str] = None
    evse: Optional[EVSEType] = None


@dataclass
class VariableType:
    """
    Reference key to a component-variable.
    VariableType is used by: ComponentVariableType,
    GetVariablesRequest.GetVariableDataType,
    GetVariablesResponse.GetVariableResultType,
    NotifyMonitoringReportRequest.MonitoringDataType,
    NotifyReportRequest.ReportDataType,
    SetVariableMonitoringRequest.SetMonitoringDataType,
    SetVariableMonitoringResponse.SetMonitoringResultType,
    SetVariablesRequest.SetVariableDataType,
    SetVariablesResponse.SetVariableResultType,
    NotifyEventRequest.EventDataType
    """

    name: str
    instance: Optional[str] = None


@dataclass
class ComponentVariableType:
    """
    Class to report components, variables and variable attributes and
    characteristics.
    ComponentVariableType is used by: GetMonitoringReportRequest,
    GetReportRequest
    """

    component: ComponentType
    variable: Optional[VariableType] = None


@dataclass
class CompositeScheduleType:
    """
    CompositeScheduleType is used by: GetCompositeScheduleResponse
    """

    evse_id: int
    duration: int
    schedule_start: str
    charging_rate_unit: enums.ChargingRateUnitType
    charging_schedule_period: ChargingSchedulePeriodType


@dataclass
class CostType:
    """
    CostType is used by: ConsumptionCostType
    """

    cost_kind: enums.CostKindType
    amount: int
    amount_multiplier: Optional[int] = None


@dataclass
class ConsumptionCostType:
    """
    ConsumptionCostType is used by: SalesTariffEntryType
    """

    start_value: float
    cost: CostType


@dataclass
class EventDataType:
    """
    Class to report an event notification for a component-variable.
    EventDataType is used by: NotifyEventRequest
    """

    event_id: int
    timestamp: str
    trigger: enums.EventTriggerType
    actual_value: str
    event_notification_type: enums.EventNotificationType
    component: ComponentType
    variable: VariableType
    cause: Optional[int] = None
    tech_code: Optional[str] = None
    cleared: Optional[bool] = None
    transaction_id: Optional[str] = None
    variable_monitoring_id: Optional[int] = None


@dataclass
class FirmwareType:
    """
    Represents a copy of the firmware that can be loaded/updated on the
    Charging Station.
    FirmwareType is used by: UpdateFirmwareRequest
    """

    location: str
    retrival_date_time: str
    install_date_time: Optional[str] = None
    signing_certificate: Optional[str] = None
    signature: Optional[str] = None


@dataclass
class GetVariableDataType:
    """
    Class to hold parameters for GetVariables request.
    GetVariableDataType is used by: GetVariablesRequest
    """

    component: ComponentType
    variable: VariableType
    attribute_type: Optional[enums.AttributeType] = None


@dataclass
class GetVariableResultType:
    """
    Class to hold results of GetVariables request.
    GetVariableResultType is used by: GetVariablesResponse
    """

    attribute_status: enums.GetVariableStatusType
    component: ComponentType
    attribute_type: Optional[enums.AttributeType] = None
    attribute_value: Optional[str] = None


@dataclass
class IdTokenType:
    """
    Contains a case insensitive identifier to use for the authorization and the
    type of authorization to support multiple forms of identifiers.
    IdTokenType is used by: AuthorizationData, Common:IdTokenInfoType,
    RequestStartTransactionRequest, AuthorizeRequest, TransactionEventRequest,
    ReserveNowRequest, CustomerInformationRequest
    """

    id_token: str
    type: enums.IdTokenType
    additonal_info: Optional[AdditionalInfoType] = None


@dataclass
class MessageContentType:
    """
    Contains message details, for a message to be displayed on a Charging
    Station.
    MessageContentType is used by: IdTokenInfoType, MessageInfoType,
    TransactionEventResponse
    """

    format: enums.MessageFormatType
    content: str
    language: Optional[str] = None


@dataclass
class IdTokenInfoType:
    """
    Contains status information about an identifier. It is advised to not stop
    charging for a token that expires during charging, as ExpiryDate is only
    used for caching purposes. If ExpiryDate is not given, the status has no
    end date.
    IdTokenInfoType is used by: AuthorizationData, AuthorizeResponse,
    TransactionEventResponse
    """

    status: enums.AuthorizationStatusType
    cache_expiry_date_time: Optional[str] = None
    charging_priortiy: Optional[int] = None
    language_1: Optional[str] = None
    evse_id: Optional[int] = None
    language_2: Optional[str] = None
    group_id_token: Optional[IdTokenType] = None
    personal_message: Optional[MessageContentType] = None


@dataclass
class AuthorizationData:
    """
    Contains the identifier to use for authorization.
    AuthorizationData is used by: SendLocalListRequest
    """

    id_token: IdTokenType
    id_token_info: Optional[IdTokenInfoType] = None


@dataclass
class LogParametersType:
    """
    Generic class for the configuration of logging entries.
    LogParametersType is used by: GetLogRequest
    """

    remote_location: str
    oldest_timestamp: Optional[str] = None
    latest_timestamp: Optional[str] = None


@dataclass
class MessageInfoType:
    """
    Contains message details, for a message to be displayed on a Charging
    Station.
    MessageInfoType is used by: SetDisplayMessageRequest,
    NotifyDisplayMessagesRequest
    """

    id: int
    priority: enums.MessagePriorityType
    message: MessageContentType
    state: Optional[enums.MessageStateType] = None
    start_date_time: Optional[str] = None
    end_data_time: Optional[str] = None
    transaction_id: Optional[str] = None
    display: Optional[ComponentType] = None


@dataclass
class SignedMeterValueType:
    """
    Represent a signed version of the meter value.
    SignedMeterValueType is used by: SampledValueType
    """

    signed_meter_data: str
    signing_method: str
    encoding_method: str
    public_key: str


@dataclass
class UnitOfMeasureType:
    """
    Represents a UnitOfMeasure with a multiplier
    UnitOfMeasureType is used by: SampledValueType
    """

    unit: Optional[str] = None
    multiplier: Optional[int] = None


@dataclass
class SampledValueType:
    """
    Single sampled value in MeterValues. Each value can be accompanied by
    optional fields.
    To save on mobile data usage, default values of all of the optional fields
    are such that. The value without any additional fields will be interpreted,
    as a register reading of active import energy in Wh (Watt-hour) units.
    SampledValueType is used by: MeterValueType
    """

    value: float
    context: Optional[enums.ReadingContextType] = None
    phase: Optional[enums.PhaseType] = None
    location: Optional[enums.LocationType] = None
    signed_meter_value: Optional[SignedMeterValueType] = None
    unit_of_measure: Optional[UnitOfMeasureType] = None


@dataclass
class MeterValueType:
    """
    Collection of one or more sampled values in MeterValuesRequest and
    TransactionEvent. All sampled values in a MeterValue are sampled at the
    same point in time.
    MeterValueType is used by: MeterValuesRequest, TransactionEventRequest
    """

    timestamp: str
    sampled_value: SampledValueType


@dataclass
class ModemType:
    """
    Defines parameters required for initiating and maintaining wireless
    communication with other devices.
    ModemType is used by: BootNotificationRequest.ChargingStationType
    """

    iccid: Optional[str] = None
    imsi: Optional[str] = None


@dataclass
class VariableMonitoringType:
    """
    A monitoring setting for a variable.
    VariableMonitoringType is used by:
    NotifyMonitoringReportRequest.MonitoringDataType
    """

    id: int
    transaction: bool
    value: float
    type: enums.MonitorType
    severity: int


@dataclass
class MonitoringDataType:
    """
    Class to hold parameters of SetVariableMonitoring request.
    MonitoringDataType is used by: NotifyMonitoringReportRequest
    """

    component: ComponentType
    variable: VariableType
    variable_monitoring: VariableMonitoringType


@dataclass
class VPNType:
    """
    VPN Configuration settings
    VPNType is used by: SetNetworkProfileRequest.NetworkConnectionProfileType
    """

    server: str
    user: str
    password: str
    key: str
    type: enums.VPNType
    group: Optional[str] = None


@dataclass
class NetworkConnectionProfileType:
    """
    The NetworkConnectionProfile defines the functional and technical
    parameters of a communication link.
    NetworkConnectionProfileType is used by: SetNetworkProfileRequest
    """

    ocpp_version: enums.OCPPVersionType
    ocpp_transport: enums.OCPPTransportType
    ocpp_csms_url: str
    message_timeout: int
    security_profile: int
    ocpp_interface: enums.OCPPInterfaceType
    vpn: Optional[VPNType] = None
    apn: Optional[APNType] = None


@dataclass
class ChargingStationType:
    """
    The physical system where an Electrical Vehicle (EV) can be charged.
    ChargingStationType is used by: BootNotificationRequest
    """

    vendor_name: str
    model: str
    modem: Optional[ModemType] = None
    serial_number: Optional[str] = None
    firmware_version: Optional[str] = None


@dataclass
class OCSPRequestDataType:
    """
    OCSPRequestDataType is used by: AuthorizeRequest,
    GetCertificateStatusRequest
    """

    hash_algorithm: enums.HashAlgorithmType
    issuer_name_hash: str
    issuer_key_hash: str
    serial_number: str
    responder_url: str


@dataclass
class VariableAttributeType:
    """
    Attribute data of a variable.
    VariableAttributeType is used by: NotifyReportRequest.ReportDataType
    """

    type: Optional[enums.AttributeType] = None
    value: Optional[str] = None
    mutability: Optional[enums.MutabilityType] = None
    persistent: Optional[bool] = None
    constant: Optional[bool] = None


@dataclass
class VariableCharacteristicsType:
    """
    Fixed read-only parameters of a variable.
    VariableCharacteristicsType is used by: NotifyReportRequest.ReportDataType
    """

    data_type: enums.DataType
    supports_monitoring: bool
    unit: Optional[str] = None
    min_limit: Optional[float] = None
    max_limit: Optional[float] = None
    values_list: Optional[str] = None


@dataclass
class ReportDataType:
    """
    Class to report components, variables and variable attributes and
    characteristics
    ReportDataType is used by: NotifyReportRequest
    """

    component: ComponentType
    variable: VariableType
    variable_attribute: VariableAttributeType
    variable_characteristics: Optional[VariableCharacteristicsType] = None


@dataclass
class SalesTariffType:
    """
    This dataType is based on dataTypes from ISO 15118-2.
    SalesTariffType is used by: ChargingScheduleType
    """

    id: int
    sales_tariff_entry: SalesTariffEntryType
    sales_tariff_description: Optional[str] = None
    num_e_price_levels: Optional[int] = None


@dataclass
class SetMonitoringDataType:
    """
    Class to hold parameters of SetVariableMonitoring request.
    SetMonitoringDataType is used by: SetVariableMonitoringRequest
    """

    value: float
    type: enums.MonitorType
    severity: int
    component: ComponentType
    variable: VariableType
    id: Optional[int] = None
    transaction: Optional[bool] = None


@dataclass
class SetMonitoringResultType:
    """
    Class to hold result of SetVariableMonitoring request.
    SetMonitoringResultType is used by: SetVariableMonitoringResponse
    """

    status: enums.SetMonitoringStatusType
    type: enums.MonitorType
    severity: int
    component: ComponentType
    variable: VariableType
    id: Optional[int] = None
    status_info: Optional[StatusInfoType] = None


@dataclass
class SetVariableDataType:
    """SetVariableDataType is used by: SetVariablesRequest"""

    attribute_value: str
    component: ComponentType
    variable: VariableType
    attribute_type: Optional[enums.AttributeType] = None


@dataclass
class SetVariableResultType:
    """SetVariableResultType is used by: SetVariablesResponse"""

    attribute_status: enums.SetVariableStatusType
    component: ComponentType
    variable: VariableType
    attribute_type: Optional[enums.AttributeType] = None
    attribute_status_info: Optional[StatusInfoType] = None


@dataclass
class TransactionType:
    """TransactionType is used by: TransactionEventRequest"""

    transaction_id: str
    charging_state: Optional[enums.ChargingStateType] = None
    time_spent_charging: Optional[int] = None
    stopped_reason: Optional[enums.ReasonType] = None
    remote_start_id: Optional[int] = None



class OCPPError(Exception):
    """ Base class for all OCPP errors. It shouldn't be raised, only it
    subclasses.
    """
    default_description = ""

    def __init__(self, description=None, details=None):
        self.description = description
        if description is None:
            self.description = self.default_description

        self.details = details
        if self.details is None:
            self.details = {}

    def __eq__(self, other):
        if other.__class__ is self.__class__:
            return (self.description, self.details) == \
                (other.description, other.details)

        return NotImplemented

    def __repr__(self):
        return f"<{self.__class__.__name__} - description={self.description},"\
               f" details={self.details}>"

    def __str__(self):
        return f"{self.__class__.__name__}: {self.description},"\
               f" {self.details}"


class NotImplementedError(OCPPError):
    code = "NotImplemented"
    default_description = "Request Action is recognized but not supported by \
                          the receiver"


class InternalError(OCPPError):
    code = "InternalError"
    default_description = "An internal error occurred and the receiver was \
                          able to process the requested Action successfully"


class ProtocolError(OCPPError):
    code = "ProtocolError"
    default_description = "Payload for Action is incomplete"


class SecurityError(OCPPError):
    code = "SecurityError"
    default_description = "During the processing of Action a security issue \
                          occurred preventing receiver from completing the \
                          Action successfully"


class FormatViolationError(OCPPError):
    code = "FormatViolation"
    default_description = "Payload for Action is syntactically incorrect or \
                          structure for Action"


class PropertyConstraintViolationError(OCPPError):
    code = "PropertyConstraintViolation"
    default_description = "Payload is syntactically correct but at least \
                          one field contains an invalid value"


class OccurenceConstraintViolationError(OCPPError):
    code = "OccurenceConstraintViolation"
    default_description = "Payload for Action is syntactically correct but \
                          at least one of the fields violates occurence \
                          constraints"


class TypeConstraintViolationError(OCPPError):
    code = "TypeConstraintViolation"
    default_description = "Payload for Action is syntactically correct but at \
                          least one of the fields violates data type \
                          constraints (e.g. “somestring”: 12)"


class GenericError(OCPPError):
    code = "GenericError"
    default_description = "Any other error not all other OCPP defined errors"


class ValidationError(Exception):
    """ ValidationError should be raised if validation a message payload fails.

    Note this isn't an official OCPP error!
    """
    pass


class UnknownCallErrorCodeError(Exception):
    """ Raised when a CALLERROR is received with unknown error code. """
    pass



from typing import Dict, List, Optional
from dataclasses import dataclass

from ocpp.v16.enums import (
    AvailabilityStatus,
    CancelReservationStatus,
    ChargingProfileStatus,
    ClearCacheStatus,
    ClearChargingProfileStatus,
    ConfigurationStatus,
    DataTransferStatus,
    GetCompositeScheduleStatus,
    RegistrationStatus,
    RemoteStartStopStatus,
    ReservationStatus,
    ResetStatus,
    TriggerMessageStatus,
    UpdateStatus,
    UnlockStatus,
    CertificateSignedStatus,
    CertificateStatus,
    DeleteCertificateStatus,
    GenericStatus,
    GetInstalledCertificateStatus,
    LogStatus,
    UpdateFirmwareStatus
)

# Most types of CALLRESULT messages can originate from only 1 source, either
# from a Charge Point or Central System, but not from both.
#
# Take for example the CALLRESULT for an Authorize action. This type of
# CALLRESULT can only be send from a Central System to Charging Station, not
# the other way around.
#
# For some types of CALLRESULT messages the opposite is true; for example for
# the CALLRESULT message for a Reset action. This can only come from a Charge
# Point to a Central System.
#
# The only CALLRESULT that can originate from both a Central System and a
# Charge Point is the CALLRESULT message for a DataTransfer.

# The now following section of classes are for CALLRESULT messages that flow
# from Central System to Charge Point.


@dataclass
class AuthorizePayload:
    id_tag_info: Dict


@dataclass
class BootNotificationPayload:
    current_time: str
    interval: int
    status: RegistrationStatus


@dataclass
class DiagnosticsStatusNotificationPayload:
    pass


@dataclass
class FirmwareStatusNotificationPayload:
    pass


@dataclass
class HeartbeatPayload:
    current_time: str


@dataclass
class LogStatusNotificationPayload:
    pass


@dataclass
class SecurityEventNotificationPayload:
    pass


@dataclass
class SignCertificatePayload:
    status: GenericStatus


@dataclass
class MeterValuesPayload:
    pass


@dataclass
class StartTransactionPayload:
    transaction_id: int
    id_tag_info: Dict


@dataclass
class StatusNotificationPayload:
    pass


@dataclass
class StopTransactionPayload:
    id_tag_info: Optional[Dict] = None


# The CALLRESULT messages that flow from Charge Point to Central System are
# listed in the bottom part of this module.


@dataclass
class CancelReservationPayload:
    status: CancelReservationStatus


@dataclass
class CertificateSignedPayload:
    status: CertificateSignedStatus


@dataclass
class ChangeAvailabilityPayload:
    status: AvailabilityStatus


@dataclass
class ChangeConfigurationPayload:
    status: ConfigurationStatus


@dataclass
class ClearCachePayload:
    status: ClearCacheStatus


@dataclass
class ClearChargingProfilePayload:
    status: ClearChargingProfileStatus


@dataclass
class DeleteCertificatePayload:
    status: DeleteCertificateStatus


@dataclass
class ExtendedTriggerMessagePayload:
    status: TriggerMessageStatus


@dataclass
class GetInstalledCertificateIdsPayload:
    status: GetInstalledCertificateStatus
    certificate_hash_data: Optional[List] = None


@dataclass
class GetCompositeSchedulePayload:
    status: GetCompositeScheduleStatus
    connector_id: Optional[int] = None
    schedule_start: Optional[str] = None
    charging_schedule: Optional[Dict] = None


@dataclass
class GetConfigurationPayload:
    configuration_key: Optional[List] = None
    unknown_key: Optional[List] = None


@dataclass
class GetDiagnosticsPayload:
    file_name: Optional[str] = None


@dataclass
class GetLocalListVersionPayload:
    list_version: int


@dataclass
class GetLogPayload:
    status: LogStatus
    filename: Optional[str] = None


@dataclass
class InstallCertificatePayload:
    status: CertificateStatus


@dataclass
class RemoteStartTransactionPayload:
    status: RemoteStartStopStatus


@dataclass
class RemoteStopTransactionPayload:
    status: RemoteStartStopStatus


@dataclass
class ReserveNowPayload:
    status: ReservationStatus


@dataclass
class ResetPayload:
    status: ResetStatus


@dataclass
class SendLocalListPayload:
    status: UpdateStatus


@dataclass
class SetChargingProfilePayload:
    status: ChargingProfileStatus


@dataclass
class SignedFirmwareStatusNotificationPayload:
    pass


@dataclass
class SignedUpdateFirmwarePayload:
    status: UpdateFirmwareStatus


@dataclass
class TriggerMessagePayload:
    status: TriggerMessageStatus


@dataclass
class UnlockConnectorPayload:
    status: UnlockStatus


@dataclass
class UpdateFirmwarePayload:
    pass


# The DataTransfer CALLRESULT can be send both from Central System as well as
# from a Charge Point.


@dataclass
class DataTransferPayload:
    status: DataTransferStatus
    data: Optional[str] = None



from typing import Dict, List, Optional
from dataclasses import dataclass


@dataclass
class AuthorizePayload:
    id_token_info: Dict
    certificate_status: Optional[str] = None


@dataclass
class BootNotificationPayload:
    current_time: str
    interval: int
    status: str
    status_info: Optional[Dict] = None


@dataclass
class CancelReservationPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class CertificateSignedPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class ChangeAvailabilityPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class ClearCachePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class ClearChargingProfilePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class ClearDisplayMessagePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class ClearVariableMonitoringPayload:
    clear_monitoring_result: List


@dataclass
class ClearedChargingLimitPayload:
    pass


@dataclass
class CostUpdatedPayload:
    pass


@dataclass
class CustomerInformationPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class DataTransferPayload:
    status: str
    status_info: Optional[Dict] = None
    data: Optional[Dict] = None


@dataclass
class DeleteCertificatePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class FirmwareStatusNotificationPayload:
    pass


@dataclass
class Get15118EVCertificatePayload:
    status: str
    exi_response: str
    status_info: Optional[Dict] = None


@dataclass
class GetBaseReportPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class GetCertificateStatusPayload:
    status: str
    status_info: Optional[Dict] = None
    ocsp_result: Optional[str] = None


@dataclass
class GetChargingProfilesPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class GetCompositeSchedulePayload:
    status: str
    status_info: Optional[Dict] = None
    schedule: Optional[Dict] = None


@dataclass
class GetDisplayMessagesPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class GetInstalledCertificateIdsPayload:
    status: str
    status_info: Optional[Dict] = None
    certificate_hash_data_chain: Optional[List] = None


@dataclass
class GetLocalListVersionPayload:
    version_number: int


@dataclass
class GetLogPayload:
    status: str
    status_info: Optional[Dict] = None
    filename: Optional[str] = None


@dataclass
class GetMonitoringReportPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class GetReportPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class GetTransactionStatusPayload:
    messages_in_queue: bool
    ongoing_indicator: Optional[bool] = None


@dataclass
class GetVariablesPayload:
    get_variable_result: List


@dataclass
class HeartbeatPayload:
    current_time: str


@dataclass
class InstallCertificatePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class LogStatusNotificationPayload:
    pass


@dataclass
class MeterValuesPayload:
    pass


@dataclass
class NotifyChargingLimitPayload:
    pass


@dataclass
class NotifyCustomerInformationPayload:
    pass


@dataclass
class NotifyDisplayMessagesPayload:
    pass


@dataclass
class NotifyEVChargingNeedsPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class NotifyEVChargingSchedulePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class NotifyEventPayload:
    pass


@dataclass
class NotifyMonitoringReportPayload:
    pass


@dataclass
class NotifyReportPayload:
    pass


@dataclass
class PublishFirmwarePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class PublishFirmwareStatusNotificationPayload:
    pass


@dataclass
class ReportChargingProfilesPayload:
    pass


@dataclass
class RequestStartTransactionPayload:
    status: str
    status_info: Optional[Dict] = None
    transaction_id: Optional[str] = None


@dataclass
class RequestStopTransactionPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class ReservationStatusUpdatePayload:
    pass


@dataclass
class ReserveNowPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class ResetPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class SecurityEventNotificationPayload:
    pass


@dataclass
class SendLocalListPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class SetChargingProfilePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class SetDisplayMessagePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class SetMonitoringBasePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class SetMonitoringLevelPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class SetNetworkProfilePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class SetVariableMonitoringPayload:
    set_monitoring_result: List


@dataclass
class SetVariablesPayload:
    set_variable_result: List


@dataclass
class SignCertificatePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class StatusNotificationPayload:
    pass


@dataclass
class TransactionEventPayload:
    total_cost: Optional[int] = None
    charging_priority: Optional[int] = None
    id_token_info: Optional[Dict] = None
    updated_personal_message: Optional[Dict] = None


@dataclass
class TriggerMessagePayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class UnlockConnectorPayload:
    status: str
    status_info: Optional[Dict] = None


@dataclass
class UnpublishFirmwarePayload:
    status: str


@dataclass
class UpdateFirmwarePayload:
    status: str
    status_info: Optional[Dict] = None



import asyncio
import inspect
import logging
import uuid
import re
import time
from dataclasses import asdict

from ocpp.routing import create_route_map
from ocpp.messages import Call, validate_payload, MessageType
from ocpp.exceptions import OCPPError, NotImplementedError
from ocpp.messages import unpack

LOGGER = logging.getLogger('ocpp')


def camel_to_snake_case(data):
    """
    Convert all keys of all dictionaries inside the given argument from
    camelCase to snake_case.

    Inspired by: https://stackoverflow.com/a/1176023/1073222

    """
    if isinstance(data, dict):
        snake_case_dict = {}
        for key, value in data.items():
            s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', key)
            key = re.sub('([a-z0-9])([A-Z])(?=\\S)', r'\1_\2', s1).lower()

            snake_case_dict[key] = camel_to_snake_case(value)

        return snake_case_dict

    if isinstance(data, list):
        snake_case_list = []
        for value in data:
            snake_case_list.append(camel_to_snake_case(value))

        return snake_case_list

    return data


def snake_to_camel_case(data):
    """
    Convert all keys of a all dictionaries inside given argument from
    snake_case to camelCase.

    Inspired by: https://stackoverflow.com/a/19053800/1073222
    """
    if isinstance(data, dict):
        camel_case_dict = {}
        for key, value in data.items():
            key = key.replace('soc', 'SoC')
            components = key.split("_")
            key = components[0] + "".join(
                x[:1].upper() + x[1:] for x in components[1:])
            camel_case_dict[key] = snake_to_camel_case(value)

        return camel_case_dict

    if isinstance(data, list):
        camel_case_list = []
        for value in data:
            camel_case_list.append(snake_to_camel_case(value))

        return camel_case_list

    return data


def remove_nones(dict_to_scan):
    dict_to_scan = {
        k: v for k, v in dict_to_scan.items()
        if v is not None
    }
    return dict_to_scan


class ChargePoint:
    """
    Base Element containing all the necessary OCPP1.6J messages for messages
    initiated and received by the Central System
    """
    def __init__(self, id, connection, response_timeout=30):
        """

        Args:

            charger_id (str): ID of the charger.
            connection: Connection to CP.
            response_timeout (int): When no response on a request is received
                within this interval, a asyncio.TimeoutError is raised.

        """
        self.id = id

        # The maximum time in seconds it may take for a CP to respond to a
        # CALL. An asyncio.TimeoutError will be raised if this limit has been
        # exceeded.
        self._response_timeout = response_timeout

        # A connection to the client. Currently this is an instance of gh
        self._connection = connection

        # A dictionary that hooks for Actions. So if the CS receives a it will
        # look up the Action into this map and execute the corresponding hooks
        # if exists.
        self.route_map = create_route_map(self)

        self._call_lock = asyncio.Lock()

        # A queue used to pass CallResults and CallErrors from
        # the self.serve() task to the self.call() task.
        self._response_queue = asyncio.Queue()

        # Function used to generate unique ids for CALLs. By default
        # uuid.uuid4() is used, but it can be changed. This is meant primarily
        # for testing purposes to have predictable unique ids.
        self._unique_id_generator = uuid.uuid4

    async def start(self):
        while True:
            message = await self._connection.recv()
            LOGGER.info('%s: receive message %s', self.id, message)

            await self.route_message(message)

    async def route_message(self, raw_msg):
        """
        Route a message received from a CP.

        If the message is a of type Call the corresponding hooks are executed.
        If the message is of type CallResult or CallError the message is passed
        to the call() function via the response_queue.
        """
        try:
            msg = unpack(raw_msg)
        except OCPPError as e:
            LOGGER.exception("Unable to parse message: '%s', it doesn't seem "
                             "to be valid OCPP: %s", raw_msg, e)
            return

        if msg.message_type_id == MessageType.Call:
            await self._handle_call(msg)

        elif msg.message_type_id in \
                [MessageType.CallResult, MessageType.CallError]:
            self._response_queue.put_nowait(msg)

    async def _handle_call(self, msg):
        """
        Execute all hooks installed for based on the Action of the message.

        First the '_on_action' hook is executed and its response is returned to
        the client. If there is no '_on_action' hook for Action in the message
        a CallError with a NotImplemtendError is returned.

        Next the '_after_action' hook is executed.

        """
        try:
            handlers = self.route_map[msg.action]
        except KeyError:
            raise NotImplementedError(f"No handler for '{msg.action}' "
                                      "registered.")

        if not handlers.get('_skip_schema_validation', False):
            validate_payload(msg, self._ocpp_version)

        # OCPP uses camelCase for the keys in the payload. It's more pythonic
        # to use snake_case for keyword arguments. Therefore the keys must be
        # 'translated'. Some examples:
        #
        # * chargePointVendor becomes charge_point_vendor
        # * firmwareVersion becomes firmwareVersion
        snake_case_payload = camel_to_snake_case(msg.payload)

        try:
            handler = handlers['_on_action']
        except KeyError:
            raise NotImplementedError(f"No handler for '{msg.action}' "
                                      "registered.")

        try:
            response = handler(**snake_case_payload)
            if inspect.isawaitable(response):
                response = await response
        except Exception as e:
            LOGGER.exception("Error while handling request '%s'", msg)
            response = msg.create_call_error(e).to_json()
            await self._send(response)

            return

        temp_response_payload = asdict(response)

        # Remove nones ensures that we strip out optional arguments
        # which were not set and have a default value of None
        response_payload = remove_nones(temp_response_payload)

        # The response payload must be 'translated' from snake_case to
        # camelCase. So:
        #
        # * charge_point_vendor becomes chargePointVendor
        # * firmware_version becomes firmwareVersion
        camel_case_payload = snake_to_camel_case(response_payload)

        response = msg.create_call_result(camel_case_payload)

        if not handlers.get('_skip_schema_validation', False):
            validate_payload(response, self._ocpp_version)

        await self._send(response.to_json())

        try:
            handler = handlers['_after_action']
            # Create task to avoid blocking when making a call inside the
            # after handler
            response = handler(**snake_case_payload)
            if inspect.isawaitable(response):
                asyncio.ensure_future(response)
        except KeyError:
            # '_on_after' hooks are not required. Therefore ignore exception
            # when no '_on_after' hook is installed.
            pass

    async def call(self, payload, suppress=True):
        """
        Send Call message to client and return payload of response.

        The given payload is transformed into a Call object by looking at the
        type of the payload. A payload of type BootNotificationPayload will
        turn in a Call with Action BootNotification, a HeartbeatPayload will
        result in a Call with Action Heartbeat etc.

        A timeout is raised when no response has arrived before expiring of
        the configured timeout.

        When waiting for a response no other Call message can be send. So this
        function will wait before response arrives or response timeout has
        expired. This is in line the OCPP specification

        Suppress is used to maintain backwards compatibility. When set to True,
        if response is a CallError, then this call will be suppressed. When
        set to False, an exception will be raised for users to handle this
        CallError.

        """
        camel_case_payload = snake_to_camel_case(asdict(payload))

        call = Call(
            unique_id=str(self._unique_id_generator()),
            action=payload.__class__.__name__[:-7],
            payload=remove_nones(camel_case_payload)
        )

        validate_payload(call, self._ocpp_version)

        # Use a lock to prevent make sure that only 1 message can be send at a
        # a time.
        async with self._call_lock:
            await self._send(call.to_json())
            try:
                response = \
                    await self._get_specific_response(call.unique_id,
                                                      self._response_timeout)
            except asyncio.TimeoutError:
                raise asyncio.TimeoutError(
                    f"Waited {self._response_timeout}s for response on "
                    f"{call.to_json()}."
                )

        if response.message_type_id == MessageType.CallError:
            LOGGER.warning("Received a CALLError: %s'", response)
            if suppress:
                return
            raise response.to_exception()
        else:
            response.action = call.action
            validate_payload(response, self._ocpp_version)

        snake_case_payload = camel_to_snake_case(response.payload)
        # Create the correct Payload instance based on the received payload. If
        # this method is called with a call.BootNotificationPayload, then it
        # will create a call_result.BootNotificationPayload. If this method is
        # called with a call.HeartbeatPayload, then it will create a
        # call_result.HeartbeatPayload etc.
        cls = getattr(self._call_result, payload.__class__.__name__)  # noqa
        return cls(**snake_case_payload)

    async def _get_specific_response(self, unique_id, timeout):
        """
        Return response with given unique ID or raise an asyncio.TimeoutError.
        """
        wait_until = time.time() + timeout
        try:
            # Wait for response of the Call message.
            response = await asyncio.wait_for(self._response_queue.get(),
                                              timeout)
        except asyncio.TimeoutError:
            raise

        if response.unique_id == unique_id:
            return response

        LOGGER.error('Ignoring response with unknown unique id: %s', response)
        timeout_left = wait_until - time.time()

        if timeout_left < 0:
            raise asyncio.TimeoutError

        return await self._get_specific_response(unique_id, timeout_left)

    async def _send(self, message):
        LOGGER.info('%s: send %s', self.id, message)
        await self._connection.send(message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ocpp', 'ocpp.v16', 'ocpp.v20', 'ocpp.v201']

package_data = \
{'': ['*'],
 'ocpp.v16': ['schemas/*'],
 'ocpp.v20': ['schemas/*'],
 'ocpp.v201': ['schemas/*']}

install_requires = \
['jsonschema>=3.0,<4.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.6,<0.7']}

setup_kwargs = {
    'name': 'ocpp',
    'version': '0.10.0',
    'description': 'Python package implementing the JSON version of the Open Charge Point Protocol (OCPP).',
    'long_description': '.. image:: https://circleci.com/gh/mobilityhouse/ocpp/tree/master.svg?style=svg\n   :target: https://circleci.com/gh/mobilityhouse/ocpp/tree/master\n\n.. image:: https://img.shields.io/pypi/pyversions/ocpp.svg\n   :target: https://pypi.org/project/ocpp/\n\n.. image:: https://img.shields.io/readthedocs/ocpp.svg\n   :target: https://ocpp.readthedocs.io/en/latest/\n\nOCPP\n----\n\nPython package implementing the JSON version of the Open Charge Point Protocol\n(OCPP). Currently OCPP 1.6 (errata v4), OCPP 2.0 and OCPP 2.0.1 (Final Version)\nare supported.\n\nYou can find the documentation on `rtd`_.\n\nInstallation\n------------\n\nYou can either the project install from Pypi:\n\n.. code-block:: bash\n\n   $ pip install ocpp\n\nOr clone the project and install it manually using:\n\n.. code-block:: bash\n\n   $ pip install .\n\nQuick start\n-----------\n\nBelow you can find examples on how to create a simple OCPP 2.0 central system as\nwell as an OCPP 2.0 charge point.\n\n.. note::\n\n   To run these examples the dependency websockets_ is required! Install it by running:\n\n   .. code-block:: bash\n\n      $ pip install websockets\n\nCentral system\n~~~~~~~~~~~~~~\n\nThe code snippet below creates a simple OCPP 2.0 central system which is able\nto handle BootNotification calls. You can find a detailed explanation of the\ncode in the `Central System documentation_`.\n\n\n.. code-block:: python\n\n    import asyncio\n    import logging\n    import websockets\n    from datetime import datetime\n\n    from ocpp.routing import on\n    from ocpp.v201 import ChargePoint as cp\n    from ocpp.v201 import call_result\n\n    logging.basicConfig(level=logging.INFO)\n\n\n    class ChargePoint(cp):\n        @on(\'BootNotification\')\n        async def on_boot_notification(self, charging_station, reason, **kwargs):\n            return call_result.BootNotificationPayload(\n                current_time=datetime.utcnow().isoformat(),\n                interval=10,\n                status=\'Accepted\'\n            )\n\n\n    async def on_connect(websocket, path):\n        """ For every new charge point that connects, create a ChargePoint\n        instance and start listening for messages.\n        """\n        try:\n            requested_protocols = websocket.request_headers[\n                \'Sec-WebSocket-Protocol\']\n        except KeyError:\n            logging.info("Client hasn\'t requested any Subprotocol. "\n                     "Closing Connection")\n        if websocket.subprotocol:\n            logging.info("Protocols Matched: %s", websocket.subprotocol)\n        else:\n            # In the websockets lib if no subprotocols are supported by the\n            # client and the server, it proceeds without a subprotocol,\n            # so we have to manually close the connection.\n            logging.warning(\'Protocols Mismatched | Expected Subprotocols: %s,\'\n                            \' but client supports  %s | Closing connection\',\n                            websocket.available_subprotocols,\n                            requested_protocols)\n            return await websocket.close()\n\n        charge_point_id = path.strip(\'/\')\n        cp = ChargePoint(charge_point_id, websocket)\n\n        await cp.start()\n\n\n    async def main():\n        server = await websockets.serve(\n            on_connect,\n            \'0.0.0.0\',\n            9000,\n            subprotocols=[\'ocpp2.0.1\']\n        )\n        logging.info("WebSocket Server Started")\n        await server.wait_closed()\n\n    if __name__ == \'__main__\':\n        asyncio.run(main())\n\nCharge point\n~~~~~~~~~~~~\n\n.. code-block:: python\n\n    import asyncio\n    import logging\n    import websockets\n\n    from ocpp.v201 import call\n    from ocpp.v201 import ChargePoint as cp\n\n    logging.basicConfig(level=logging.INFO)\n\n\n    class ChargePoint(cp):\n\n       async def send_boot_notification(self):\n           request = call.BootNotificationPayload(\n                   charging_station={\n                       \'model\': \'Wallbox XYZ\',\n                       \'vendor_name\': \'anewone\'\n                   },\n                   reason="PowerUp"\n           )\n           response = await self.call(request)\n\n           if response.status == \'Accepted\':\n               print("Connected to central system.")\n\n\n    async def main():\n       async with websockets.connect(\n           \'ws://localhost:9000/CP_1\',\n            subprotocols=[\'ocpp2.0.1\']\n       ) as ws:\n\n           cp = ChargePoint(\'CP_1\', ws)\n\n           await asyncio.gather(cp.start(), cp.send_boot_notification())\n\n\n    if __name__ == \'__main__\':\n       asyncio.run(main())\n\nDebugging\n---------\n\nPython\'s default log level is `logging.WARNING`. As result most of the logs\ngenerated by this package are discarded. To see the log output of this package\nlower the log level to `logging.DEBUG`.\n\n.. code-block:: python\n\n  import logging\n  logging.basicConfig(level=logging.DEBUG)\n\nHowever, this approach defines the log level for the complete logging system.\nIn other words: the log level of all dependencies is set to `logging.DEBUG`.\n\nTo lower the logs for this package only use the following code:\n\n.. code-block:: python\n\n  import logging\n  logging.getLogger(\'ocpp\').setLevel(level=logging.DEBUG)\n  logging.getLogger(\'ocpp\').addHandler(logging.StreamHandler())\n\nLicense\n-------\n\nExcept from the documents in `docs/v16` and `docs/v201` everything is licensed under MIT_.\n© `The Mobility House`_\n\nThe documents in `docs/v16` and `docs/v201` are licensed under Creative Commons\nAttribution-NoDerivatives 4.0 International Public License.\n\n.. _Central System documentation: https://ocpp.readthedocs.io/en/latest/central_system.html\n.. _MIT: https://github.com/mobilityhouse/ocpp/blob/master/LICENSE\n.. _rtd: https://ocpp.readthedocs.io/en/latest/index.html\n.. _The Mobility House: https://www.mobilityhouse.com/int_en/\n.. _websockets: https://pypi.org/project/websockets/\n',
    'author': 'Andre Duarte',
    'author_email': 'andre.duarte@mobilityhouse.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mobilityhouse/ocpp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
